"""This module implements :py:class:`.BasePlotCreator`, the base class for plot
creators.

The interface is defined as an abstract base class and partly implemented by
the :py:class:`.BasePlotCreator`, which is no longer abstract but has only the
functionality that is general enough for all derived creators to profit from.
"""

import contextlib
import copy
import gc
import importlib
import importlib.util
import logging
import os
import time
from typing import Callable, Dict, Sequence, Tuple, Union

from paramspace import ParamSpace

from ..._copy import _deepcopy
from ..._dag_utils import resolve_placeholders as _resolve_placeholders
from ..._hash import _hash
from ...abc import AbstractPlotCreator
from ...dag import TransformationDAG
from ...data_mngr import DataManager
from ...exceptions import PlotCreatorError, SkipPlot
from ...tools import format_time as _format_time
from ...tools import recursive_update

log = logging.getLogger(__name__)

_fmt_time = lambda t: _format_time(t, ms_precision=1)

_DAG_OBJECT_CACHE: Dict[str, TransformationDAG] = dict()
"""A dict holding the previously-used :py:class:`~dantro.dag.TransformationDAG`
objects to allow re-using them in another plot function.
The keys are hashes of the configuration used in creating the DAG."""


# -----------------------------------------------------------------------------


class BasePlotCreator(AbstractPlotCreator):
    """The base class for plot creators.

    It provides the following functionality:

    - Resolving a plot function, which can be a directly given callable, an
      importable module and name, or a path to a Python file that is to be
      imported.
    - Parsing plot configuration arguments.
    - Optionally, performing data selection from the associated
      :py:class:`~dantro.data_mngr.DataManager` using the
      :ref:`data transformation framework <dag_framework>`.
    - Invoking the plot function.

    As such, the this base class is agnostic to the exact way of how plot
    output is generated; the plot function is responsible for that.
    """

    EXTENSIONS: Union[Tuple[str], str] = "all"
    """A tuple of supported file extensions.
    If ``all``, no checks for the extensions are performed."""

    DEFAULT_EXT = None
    """The default file extension to use; is only used if no default extension
    is specified during initialization"""

    DEFAULT_EXT_REQUIRED: bool = True
    """Whether a default extension is required or not. If True and the
    ``default_ext`` property evaluates to False, an error will be raised."""

    POSTPONE_PATH_PREPARATION: bool = False
    """Whether to prepare paths in the base class's
    :py:meth:`~dantro.plot.creators.base.BasePlotCreator.__call__` method
    or not. If the derived class wants to take care of this on their own, this
    should be set to True and the
    :py:meth:`~dantro.plot.creators.base.BasePlotCreator._prepare_path`
    method, adjusted or not, should be called at another point of the plot
    execution."""

    OUT_PATH_EXIST_OK: bool = False
    """Whether a warning should be shown (instead of an error), when a plot
    file already exists at the specified output path"""

    DAG_USE_DEFAULT: bool = False
    """Whether the :ref:`data transformation framework <pcr_base_dag_support>`
    is enabled *by default*; this can still be controlled by the ``use_dag``
    argument of the plot configuration.
    """

    DAG_RESOLVE_PLACEHOLDERS: bool = True
    """Whether placeholders in the plot config,
    :py:class:`~dantro._dag_utils.ResultPlaceholder` objects, should be
    replaced with results from the data transformations."""

    DAG_TO_KWARG_MAPPING: Dict[str, str] = {
        "results_dict": "data",
        "dag_object": "dag",
    }
    """The keyword argument names by which to pass the data transformation
    results (``results_dict``) or the :py:class:`~dantro.dag.TransformationDAG`
    object itself (``dag_object``) to the plot function.
    """

    # .........................................................................

    def __init__(
        self,
        name: str,
        *,
        dm: DataManager,
        plot_func: Callable,
        default_ext: str = None,
        exist_ok: Union[bool, str] = None,
        raise_exc: bool = None,
        **plot_cfg,
    ):
        """Create a plot creator instance for a plot with the given ``name``.

        Typically, a creator has not be instantiated separately, but the
        :py:class:`~dantro.plot_mngr.PlotManager` takes care of it.

        Args:
            name (str): The name of this plot
            dm (DataManager): The data manager that contains the data to plot
            default_ext (str, optional): The default extension to use; needs
                to be in ``EXTENSIONS``, if that class variable is not set to
                'all'. The value given here is needed by the PlotManager to
                build the output path.
            exist_ok (Union[bool, str], optional): If True, no error will be
                raised when a plot already exists at the specified output path.
                If None, the value specified in the ``OUT_PATH_EXIST_OK`` class
                variable will be used to determine this behaviour.
                If ``skip``, will skip the plot, allowing other plots to be
                carried out; see :ref:`plot_mngr_skipping_plots`.
            raise_exc (bool, optional): Whether to raise exceptions during the
                plot procedure. This does not pertain to *all* exceptions, but
                only to those that would *unnecessarily* stop plotting.
                Furthermore, whether this setting is used or not depends on the
                used creator specialization.
            **plot_cfg: The default configuration for the plot(s) that this
                creator is supposed to create.

        Raises:
            ValueError: On bad ``base_module_file_dir`` or ``default_ext``
        """
        self._name = name
        self._dm = dm
        self._plot_cfg = plot_cfg
        self._plot_func = plot_func
        self._using_dag = None
        self._dag_obj_cache = _DAG_OBJECT_CACHE
        self._out_path = None
        self._exist_ok = (
            self.OUT_PATH_EXIST_OK if exist_ok is None else exist_ok
        )
        self.raise_exc = raise_exc

        # Property-managed attributes
        self._logstr = None
        self._default_ext = None
        self._dag = None

        # DAG visualization
        self._dag_vis_kwargs = None
        self._dag_vis_done_for = []

        # Set the default extension, first from argument, then default.
        # Then check that it was set correctly
        if default_ext is not None:
            self.default_ext = default_ext

        elif self.DEFAULT_EXT is not None:
            self.default_ext = self.DEFAULT_EXT

        if self.DEFAULT_EXT_REQUIRED and not self.default_ext:
            raise ValueError(
                f"{self.logstr} requires a default extension, but neither "
                f"the argument ('{default_ext}') nor the DEFAULT_EXT class "
                f"variable ('{self.DEFAULT_EXT}') was set."
            )

    # .. Properties ...........................................................

    @property
    def name(self) -> str:
        """Returns this creator's name"""
        return self._name

    @property
    def classname(self) -> str:
        """Returns this creator's class name"""
        return self.__class__.__name__

    @property
    def logstr(self) -> str:
        """Returns the classname and name of this object; a combination often
        used in logging..."""
        if not self._logstr:
            self._logstr = f"{self.classname} for '{self.name}'"
        return self._logstr

    @property
    def dm(self) -> DataManager:
        """Return the DataManager"""
        return self._dm

    @property
    def plot_func(self) -> Callable:
        """Returns the plot function"""
        return self._plot_func

    @property
    def plot_func_name(self) -> str:
        """Returns a readable name of the plot function"""
        pf = self.plot_func
        return getattr(pf, "name", pf.__name__)

    @property
    def plot_cfg(self) -> Dict[str, dict]:
        """Returns a deepcopy of the plot configuration, assuring that plot
        configurations are completely independent of each other.
        """
        return copy.deepcopy(self._plot_cfg)

    @property
    def default_ext(self) -> str:
        """Returns the default extension to use for the plots"""
        return self._default_ext

    @default_ext.setter
    def default_ext(self, val: str) -> None:
        """Sets the default extension.

        Unless :py:attr:`.EXTENSIONS` is set to ``all``, needs to be a valid
        extension.
        """
        if self.EXTENSIONS != "all" and val not in self.EXTENSIONS:
            raise ValueError(
                f"Extension '{val}' not supported in {self.logstr}. "
                f"Supported extensions are: {', '.join(self.EXTENSIONS)}"
            )

        self._default_ext = val

    @property
    def dag(self) -> TransformationDAG:
        """The associated TransformationDAG object. If not set up, raises."""
        if self._dag is not None:
            return self._dag
        raise ValueError(
            f"{self.logstr} has no TransformationDAG associated (yet)!"
        )

    # .. Main API functions, required by PlotManager ..........................

    def __call__(self, *, out_path: str, **update_plot_cfg):
        """Perform the plot, updating the configuration passed to __init__
        with the given values and then calling :py:meth:`.plot`.

        Args:
            out_path (str): The full output path to store the plot at
            **update_plot_cfg: Keys with which to update the default plot
                configuration

        Returns:
            The return value of the :py:meth:`.plot` method, which is an
            abstract method in
            :py:class:`~dantro.plot.creators.base.BasePlotCreator`.
        """
        # Get (a deep copy of) the initial plot config, update it with new args
        cfg = self.plot_cfg
        if update_plot_cfg:
            cfg = recursive_update(cfg, copy.deepcopy(update_plot_cfg))

        # Allow derived creators to check whether this plot should be skipped
        self._check_skipping(plot_kwargs=cfg)

        # Find out if it's ok if out_path already exists, then prepare the path
        exist_ok = self._exist_ok
        if "exist_ok" in cfg:
            exist_ok = cfg.pop("exist_ok")

        if not self.POSTPONE_PATH_PREPARATION:
            self._prepare_path(out_path, exist_ok=exist_ok)

        # Now call the plotting function with these arguments
        return self.plot(out_path=out_path, **cfg)

    def plot(
        self,
        *,
        out_path: str,
        use_dag: bool = None,
        **func_kwargs,
    ):
        """Prepares argument for the plot function and invokes it.

        Args:
            out_path (str): The output path for the resulting file
            use_dag (bool, optional): Whether to use the :ref:`dag_framework`
                to select and transform data that can be used in the plotting
                function. If not given, will query the plot function attributes
                for whether the DAG should be used. If not, the data selection
                has to occur separately inside the plot function. Note that
                this may require a different plot function signature.
            **func_kwargs: Passed on to the plot function
        """
        # Store the output path, needed by downstream methods
        self._out_path = out_path

        # Prepare arguments, also performing plot data selection
        args, kwargs = self._prepare_plot_func_args(
            use_dag=use_dag, out_path=out_path, **func_kwargs
        )

        # Call the plot function, optionally generating a DAG visualization
        self._invoke_plot_func(*args, **kwargs)

    def get_ext(self) -> str:
        """Returns the extension to use for the upcoming plot by checking
        the supported extensions and can be subclassed to have different
        behaviour.
        """
        return self.default_ext

    def prepare_cfg(
        self, *, plot_cfg: dict, pspace: Union[ParamSpace, dict]
    ) -> Tuple[dict, ParamSpace]:
        """Prepares the plot configuration for the
        :py:class:`~dantro.plot_mngr.PlotManager`. This function is called by
        the manager before the first plot is to be created.

        The base implementation just passes the given arguments through.
        However, it can be re-implemented by derived classes to change the
        behaviour of the plot manager, e.g. by converting a plot configuration
        to a parameter space.
        """
        return plot_cfg, pspace

    # .........................................................................

    def _prepare_path(
        self, out_path: str, *, exist_ok: Union[bool, str]
    ) -> None:
        """Prepares the output path, creating directories if needed, then
        returning the full absolute path.

        This is called from :py:meth:`.__call__` and is meant to postpone
        directory creation as far as possible.

        Args:
            out_path (str): The absolute output path to start with
            exist_ok (Union[bool, str]): If False, will raise if a file of that
                name already exists; if True, will emit a warning instead.
                If ``'skip'``, will initiate skipping of this plot.

        Raises:
            FileExistsError: Raised on already existing out path and exist_ok
                being False.
        """
        if os.path.exists(out_path):
            msg = (
                "There already exists a file at the specified output path "
                f"for {self.logstr}:\n  {out_path}"
            )
            if not exist_ok:
                raise FileExistsError(msg)
            elif exist_ok == "skip":
                raise SkipPlot(f"Plot output already exists:\n  {out_path}")
            else:
                log.warning(msg)

        # Ensure that all necessary directories exist
        os.makedirs(os.path.dirname(out_path), exist_ok=True)

    def _check_skipping(self, *, plot_kwargs: dict):
        """A method that can be specialized by derived plot creators to check
        whether a plot should be skipped.
        Is invoked from the :py:meth:`.__call__` method, *after*
        :py:meth:`._perform_data_selection` (for plots with activated data
        selection via DAG), and *prior to* :py:meth:`._prepare_path`
        (such that path creation can be avoided).

        In cases where this plot is to be skipped, the custom exception
        :py:class:`~dantro.exceptions.SkipPlot` should be raised,
        the error message allowing to specify a reason for skipping the plot.

        .. note::

            While the base class method may be a no-op, it should still be
            called via ``super()._check_skipping`` from the derived classes.

        Args:
            plot_kwargs (dict): The full plot configuration
        """
        pass

    # .........................................................................
    # Plot argument preparation and function invocation

    def _prepare_plot_func_args(
        self, *, use_dag: bool = None, **kwargs
    ) -> Tuple[tuple, dict]:
        """Prepares the arguments passed to the plot function.

        The passed keyword arguments are carried over; no positional arguments
        are possible.
        Subsequently, possible signatures look as follows:

        - When using the data transformation framework, there are *no*
          positional arguments.
        - When *not* using the data transformation framework, the *only*
          positional argument is the :py:class:`~dantro.data_mngr.DataManager`
          instance that is associated with this plot.

        .. note::

            When subclassing this function, the parent method (this one) should
            still be called to maintain base functionality.

        Args:
            use_dag (bool, optional): Whether to use the data transformation
                framework
            **kwargs: Additional kwargs

        Returns:
            Tuple[tuple, dict]: an (empty) tuple of positional arguments and a
                dict of keyword arguments.
        """
        # Perform data selection and transformation, if the plot creator class
        # supports it.
        # Even if the creator supports it, it might be disabled in the config;
        # in that case, the method below behaves like a passthrough of the cfg,
        # filtering out all transformation-related arguments.
        # The returned kwargs are the adjusted plot function keyword arguments.
        using_dag, kwargs = self._perform_data_selection(
            use_dag=use_dag, plot_kwargs=kwargs
        )
        self._using_dag = using_dag

        # Aggregate as (args, kwargs), passed on to plot function. When using
        # the DAG, the DataManager is NOT passed along, as it is accessible via
        # the tags of the DAG.
        if not using_dag:
            return ((self.dm,), kwargs)
        return ((), kwargs)

    def _invoke_plot_func(self, *args, **kwargs):
        """Method that invokes the plot function with the prepared arguments.

        This additionally allows to generate a DAG visualization in case the
        plotting failed or succeeded.
        """
        log.info("Now calling plotting function '%s' ...", self.plot_func_name)
        try:
            self.plot_func(*args, **kwargs)

        except:
            if self._using_dag:
                self._generate_DAG_vis(
                    scenario="plot_error", **self._dag_vis_kwargs
                )
            raise

        else:
            if self._using_dag:
                self._generate_DAG_vis(
                    scenario="plot_success", **self._dag_vis_kwargs
                )

        log.note("Plotting function '%s' returned.", self.plot_func_name)

    # .........................................................................
    # Data selection interface, using TransformationDAG

    def _perform_data_selection(
        self, *, use_dag: bool = None, plot_kwargs: dict, **shared_kwargs
    ) -> Tuple[bool, dict]:
        """If this plot creator supports data selection and transformation, it
        is carried out in this method.

        This method uses a number of other private methods to carry out the
        setup of the DAG, computing it and combining its results with the
        remaining plot configuration. Those methods have access to a subset of
        the whole configuration, thus allowing to parse the parameters that
        they need.

        This method also sets the ``_dag`` attribute, making the created
        :py:class:`~dantro.dag.TransformationDAG` object available for further
        processing downstream.

        Furthermore, this method invokes placeholder resolution by applying
        :py:func:`~dantro._dag_utils.resolve_placeholders` on the plot config.

        .. note::

            For specializing the behaviour of the data selection and transform,
            it is best to specialize *NOT* this method, but the more granular
            DAG-related private methods.

        .. warning::

            If subclassing this method, make sure to either invoke this parent
            method or set the ``_dag`` attribute in the subclass's method.
            Also note that, when subclassing, the ability to resolve the
            placeholders gets lost / has to be re-implemented in the subclass.

        Args:
            use_dag (bool, optional): The main toggle for whether the DAG
                should be used or not. This is passed as default value to
                another method, which takes the final decision on whether the
                DAG is used or not. If None, will first inspect whether the
                plot function declared that the DAG is to be used.
                If still None, will NOT use the DAG.
            plot_kwargs (dict): The plot configuration
            **shared_kwargs: Shared keyword arguments that are passed through
                to the helper methods :py:meth:`._use_dag` and
                :py:meth:`._get_dag_params`.

        Returns:
            Tuple[bool, dict]: Whether data selection was used and the plot
                configuration that can be passed on to the main ``plot``
                method.
        """
        # Determine whether the DAG framework should be used or not
        if not self._use_dag(
            use_dag=use_dag, plot_kwargs=plot_kwargs, **shared_kwargs
        ):
            # Only return the plot configuration, without DAG-related keys
            return False, plot_kwargs
        # else: DAG should be used
        # Extract DAG-related parameters from the plot configuration. These are
        # not available in the plotting function.
        dag_params, plot_kwargs = self._get_dag_params(
            **plot_kwargs, **shared_kwargs
        )

        # Create the DAG object, optionally reading from and/or writing to the
        # DAG object cache. Then make available to other parts.
        dag = self._setup_dag(dag_params["init"], **dag_params["cache"])
        self._dag = dag

        # Then compute results
        dag_results = self._compute_dag(dag, **dag_params["compute"])

        # If enabled, perform placeholder resolution in plot_kwargs
        if self.DAG_RESOLVE_PLACEHOLDERS:
            plot_kwargs = _resolve_placeholders(plot_kwargs, dag=dag)

        # Prepare the parameters passed back to __call__ and on to self.plot
        kws = self._combine_dag_results_and_plot_cfg(
            dag=dag,
            dag_results=dag_results,
            dag_params=dag_params,
            plot_kwargs=plot_kwargs,
        )
        return True, kws

    def _use_dag(self, *, use_dag: bool, plot_kwargs: dict) -> bool:
        """Whether the DAG should be used or not. This method extends that of
        the base class by additionally checking the plot function attributes
        for any information regarding the DAG.

        This relies on the
        :py:class:`~dantro.plot.utils.plot_func.is_plot_func`
        decorator to set a number of function attributes.
        """
        # If None was given, check the plot function attributes
        if use_dag is None:
            use_dag = getattr(self.plot_func, "use_dag", None)

        # If still None, default to the class variable default
        if use_dag is None:
            use_dag = self.DAG_USE_DEFAULT

        # Complain, if tags were required, but DAG usage was disabled
        if not use_dag and getattr(self.plot_func, "required_dag_tags", None):
            raise ValueError(
                f"The plot function {self.plot_func} requires DAG tags to be "
                "computed, but DAG usage was disabled."
            )

        return use_dag

    def _get_dag_params(
        self,
        *,
        select: dict = None,
        transform: Sequence[dict] = None,
        compute_only: Sequence[str] = None,
        dag_options: dict = None,
        dag_object_cache: dict = None,
        dag_visualization: dict = None,
        invocation_options: dict = None,
        **plot_kwargs,
    ) -> Tuple[dict, dict]:
        """Filters out and parses parameters that are needed for initialization
        of the :py:class:`~dantro.dag.TransformationDAG` in
        :py:meth:`._setup_dag` and computation in :py:meth:`_compute_dag`.

        Args:
            select (dict, optional): DAG selection
            transform (Sequence[dict], optional): DAG transformation
            compute_only (Sequence[str], optional): DAG tags to be computed
            dag_options (dict, optional): Other DAG options for initialization
            dag_object_cache (dict, optional): Cache options for the DAG object
                itself. Expected keys are ``read``, ``write``, ``clear``.
            dag_visualization (dict, optional): If given, controls whether the
                DAG used for data transformations should also be plotted, e.g.
                to make debugging easier.
            invocation_options (dict, optional): Controls whether to pass
                certain objects on to the plot functio or not. Supported keys
                are ``pass_dag_object_along`` and ``unpack_dag_results``, which
                take precedence over the plot function attributes of the same
                name which are set by the plot function decorator
                :py:class:`~dantro.plot.utils.plot_func.is_plot_func`.
            **plot_kwargs: The remaining plot configuration

        Returns:
            Tuple[dict, dict]: Tuple of DAG parameter dict and plot kwargs
        """
        # Top-level arguments
        init_kwargs = dict(select=select, transform=transform)
        compute_kwargs = dict(compute_only=compute_only)
        cache_kwargs = dag_object_cache if dag_object_cache else {}
        vis_kwargs = dag_visualization if dag_visualization else {}

        # Options. Only add those, if available
        dag_options = dag_options if dag_options else {}
        if dag_options:
            init_kwargs = dict(**init_kwargs, **dag_options)

        dag_params = dict(
            init=init_kwargs,
            compute=compute_kwargs,
            cache=cache_kwargs,
            visualization=vis_kwargs,
        )

        # Also store visualization kwargs as attribute
        self._dag_vis_kwargs = vis_kwargs
        self._dag_vis_done_for = []

        # Determine whether the DAG object should be passed along to the func
        invocation_options = invocation_options if invocation_options else {}
        _pass_dag = invocation_options.get("pass_dag_object_along")
        if _pass_dag is None:
            _pass_dag = getattr(self.plot_func, "pass_dag_object_along", False)
        dag_params["pass_dag_object_along"] = _pass_dag

        # Determine whether the DAG results should be unpacked when passing
        # them to the plot function
        _unpack = invocation_options.get("unpack_dag_results")
        if _unpack is None:
            _unpack = getattr(self.plot_func, "unpack_dag_results", False)
        dag_params["unpack_dag_results"] = _unpack

        return dag_params, plot_kwargs

    def _setup_dag(
        self,
        init_params: dict,
        *,
        read: bool = False,
        write: bool = False,
        clear: bool = False,
        collect_garbage: bool = None,
        use_copy: bool = True,
    ) -> TransformationDAG:
        """Creates a :py:class:`~dantro.dag.TransformationDAG` object from the
        given initialization parameters.
        Optionally, will use a hash of the initialization parameters to reuse
        a deep copy of a cached object.

        In case no cached version was available or caching was disabled, uses
        :py:meth:`._create_dag` to create the object.

        Args:
            init_params (dict): Initialization parameters, passed on to the
                ``_create_dag`` method.
            read (bool, optional): Whether to read from memory cache
            write (bool, optional): Whether to write to memory cache
            clear (bool, optional): Whether to clear the whole memory cache,
                can be useful if many objects were stored and memory runs low.
                Afterwards, garbage collection may be required to actually free
                the memory; see ``collect_garbage``.
            collect_garbage (bool, optional): If True, will invoke garbage
                collection; this may be required after clearing the cache but
                may also be useful to invoke separately from that.
                If None, will invoke garbage collection automatically if the
                cache was set to be cleared.
            use_copy (bool, optional): Whether to work on a (deep) copy of the
                cached DAG object. This reduces memory footprint, but may not
                bring a noticeable speedup.
        """
        t0 = time.time()

        log.note("Setting up data transformation framework ...")

        dag = None
        cp_func = _deepcopy if use_copy else lambda d: d

        # Compute the cache key only once and only if needed
        if read or write:
            cache_key = _hash(repr(init_params))

        if read:
            dag = cp_func(self._dag_obj_cache.get(cache_key))

        if dag is not None:
            log.remark(
                "Loaded TransformationDAG from memory cache. "
                "(copy? %s, cache size: %d)",
                use_copy,
                len(self._dag_obj_cache),
            )

        else:
            dag = self._create_dag(**init_params)

            if write and cache_key not in self._dag_obj_cache:
                self._dag_obj_cache[cache_key] = cp_func(dag)
                log.remark(
                    "Stored TransformationDAG in memory cache. "
                    "(copy? %s, cache size: %d)",
                    use_copy,
                    len(self._dag_obj_cache),
                )

        if clear:
            self._dag_obj_cache.clear()
            log.remark("TransformationDAG memory cache cleared.")

        if collect_garbage or (collect_garbage is None and clear):
            log.remark(
                "Invoking garbage collection ... (this may take a while)"
            )
            gc.collect()
            log.remark("Garbage collected.")

        elif collect_garbage is False:
            log.remark("NOT invoking garbage collection.")

        log.note(
            "TransformationDAG with %d nodes set up in %s.",
            len(dag.nodes),
            _fmt_time(time.time() - t0),
        )
        return dag

    def _create_dag(self, **dag_params) -> TransformationDAG:
        """Creates the actual DAG object"""
        return TransformationDAG(dm=self.dm, **dag_params)

    def _compute_dag(
        self,
        dag: TransformationDAG,
        *,
        compute_only: Sequence[str],
        **compute_kwargs,
    ) -> dict:
        """Compute the dag results.

        This checks whether all required tags (set by the
        :py:class:`~dantro.plot.utils.plot_func.is_plot_func` decorator)
        are set to be computed.
        """
        # Extract the required tags from the plot function attributes
        required_tags = getattr(self.plot_func, "required_dag_tags", None)

        # Make sure that all required tags are actually defined
        if required_tags:
            missing_tags = [t for t in required_tags if t not in dag.tags]

            if missing_tags:
                raise ValueError(
                    "Plot function {} required tags that were "
                    "not specified in the DAG: {}. Available "
                    "tags: {}. Please adjust the DAG "
                    "specification accordingly."
                    "".format(
                        self.plot_func_name,
                        ", ".join(missing_tags),
                        ", ".join(dag.tags),
                    )
                )

        # If the compute_only argument was not explicitly given, determine
        # whether to compute only the required tags
        if (
            compute_only is None
            and required_tags is not None
            and getattr(
                self.plot_func, "compute_only_required_dag_tags", False
            )
        ):
            log.remark(
                "Tags that are required by the plot function:  %s",
                ", ".join(required_tags),
            )
            compute_only = required_tags

        # Make sure the compute_only argument contains all the required tags
        elif compute_only is not None and required_tags is not None:
            missing_tags = [t for t in required_tags if t not in compute_only]

            if missing_tags:
                raise ValueError(
                    "Plot function {} required tags that were "
                    "not set to be computed by the DAG: {}. Make "
                    "sure to set the `compute_only` argument "
                    "such that results for all required tags "
                    "({}) will actually be computed.\n"
                    "Available tags:  {}\n"
                    "compute_only:    {}"
                    "".format(
                        self.plot_func_name,
                        ", ".join(missing_tags),
                        ", ".join(required_tags),
                        ", ".join(dag.tags),
                        ", ".join(compute_only),
                    )
                )

        # Now compute the results
        log.info("Computing data transformation results ...")
        try:
            results = dag.compute(compute_only=compute_only, **compute_kwargs)

        except:
            self._generate_DAG_vis(
                scenario="compute_error", **self._dag_vis_kwargs
            )
            raise

        else:
            log.remark("Finished computing data transformation results.")
            self._generate_DAG_vis(
                scenario="compute_success", **self._dag_vis_kwargs
            )

        return results

    def _combine_dag_results_and_plot_cfg(
        self,
        *,
        dag: TransformationDAG,
        dag_results: dict,
        dag_params: dict,
        plot_kwargs: dict,
    ) -> dict:
        """Returns a dict of plot configuration and ``data``, where all the
        DAG results are stored in.
        In case where the DAG results are to be unpacked, the DAG results will
        be made available as separate keyword arguments instead of as the
        single ``data`` keyword argument.

        Furthermore, if the plot function specified in its attributes that the
        DAG object is to be passed along, this is the place where it is
        included or excluded from the arguments.
        """
        if dag_params["unpack_dag_results"]:
            # Unpack the results such that they can be specified in the plot
            # function signature
            try:
                cfg = dict(**dag_results, **plot_kwargs)

            except TypeError as err:
                raise TypeError(
                    "Failed unpacking DAG results! There were arguments of "
                    "the same names as some DAG tags given in the plot "
                    "configuration. Make sure they have unique names or "
                    "disable unpacking of the DAG results.\n"
                    "  Keys in DAG results: {}\n"
                    "  Keys in plot config: {}\n"
                    "".format(
                        ", ".join(dag_results.keys()),
                        ", ".join(plot_kwargs.keys()),
                    )
                ) from err

        else:
            # Make the DAG results available as `data` kwarg
            cfg = dict(**plot_kwargs)
            cfg[self.DAG_TO_KWARG_MAPPING["results_dict"]] = dag_results

        # Add the `dag` kwarg, if configured to do so.
        if dag_params["pass_dag_object_along"]:
            cfg[self.DAG_TO_KWARG_MAPPING["dag_object"]] = dag

        return cfg

    # .........................................................................
    # DAG Visualization

    def _generate_DAG_vis(
        self,
        *,
        scenario: str,
        enabled: bool = True,
        plot_enabled: bool = True,
        export_enabled: bool = True,
        raise_exc: bool = None,
        when: dict = None,
        output: dict = None,
        export: dict = None,
        generation: dict = None,
        **plot_kwargs,
    ) -> Union["networkx.DiGraph", None]:
        """Generates a DAG representation according to certain criteria and
        using :py:meth:`~dantro.dag.TransformationDAG.generate_nx_graph`,
        then invokes :py:meth:`~dantro.dag.TransformationDAG.visualize` to
        create the actual visualization output.

        This method also allows to export the DAG representation using
        :py:func:`~dantro.utils.nx.export_graph`, which can then be used for
        externally working with the DAG representation.

        Also see :ref:`plot_creator_dag_vis` and :ref:`dag_graph_vis`.

        Args:
            scenario (str): The scenario in which the generation is invoked;
                this is used to describe the context in which this method was
                invoked and also becomes part of the output file name.
                See ``when`` for scenarios with certain names. If you want to
                use a different name, make sure to set ``when.always``, such
                that no scenario lookup occurs.
            enabled (bool, optional): If False, will return None.
            plot_enabled (bool, optional): Whether plotting is enabled. The
                result of the ``when`` evaluation overrules this.
            export_enabled (bool, optional): Whether exporting is enabled. The
                result of the ``when`` evaluation overrules this.
            raise_exc (bool, optional): Whether to raise exceptions if anything
                goes wrong within this method. If None, will behave in the same
                way as the creator does. For example, if set to False, an
                error in generating a DAG representation will *not* lead to an
                error being raised.
            when (dict, optional): A dict that specifies in which situations
                the generation should actually be carried out. May contain the
                following keys:

                    - ``always``: Will always generate output.
                    - ``only_once``: If True, will only generate output from
                      one scenario, skipping further invocations.
                    - ``on_compute_error``, ``on_compute_success``, and
                      ``on_plot_error``, ``on_plot_success``: Will generate
                      output only in certain named **scenarios**.
                      These can be a boolean toggle or ``'debug'`` in which
                      case the **creator's** debug flag decides whether output
                      is generated for that scenario.
                      Note that the ``raise_exc`` *argument* does not play a
                      role for that!

            output (dict, optional): A dict specifying where the DAG plot and
                potential exported files are written to. Allowed keys are:

                    - ``plot_dir``: If None, will write output aside the
                      plot output itself. Can also be an absolute path.
                    - ``path_fstr``: A format string that specifies the actual
                      output path and should/can contain the keys ``plot_dir``,
                      ``name``, and ``scenario``.

            export (dict, optional): Export specification, using networkx's
                write methods. Possible keys:

                    - ``manipulate_attrs``: Dict that controls manipulation of
                      node or edge attributes, sometimes necessary for export.
                      These are passed to
                      :py:func:`~dantro.utils.nx.manipulate_attributes`.
                    - any further keyword arguments define the output formats
                      that are to be used.
                      They can be of type ``Dict[str, Union[bool, dict]]``,
                      where the string should correspond to the name of a
                      networkx writer method. The boolean is used to enable
                      or disable a writer. If a dict is given, its content is
                      passed to the writer method.
                      Also see :py:func:`~dantro.utils.nx.export_graph`, where
                      this is implemented.

            generation (dict, optional): Graph generation arguments passed to
                :py:meth:`~dantro.dag.TransformationDAG.generate_nx_graph`.
            **plot_kwargs: Plotting-related arguments, passed on to
                :py:meth:`~dantro.dag.TransformationDAG.visualize`.

        Returns:
            Union[networkx.DiGraph, None]: Either the generated graph object
                or None, if not enabled or ``when`` was evaluated to not
                generating a DAG representation.
        """

        def should_plot(
            scenario: str,
            *,
            enabled: bool,
            always: bool = False,
            only_once: bool = False,
            on_compute_error: Union[bool, str] = True,
            on_compute_success: Union[bool, str] = False,
            on_plot_error: Union[bool, str] = False,
            on_plot_success: Union[bool, str] = False,
        ) -> bool:
            """Decides whether a DAG visualization should be created in a
            certain scenario.
            """
            if not enabled:
                return False

            if always:
                return True

            if only_once and scenario in self._dag_vis_done_for:
                return False

            scenarios = dict(
                compute_error=on_compute_error,
                compute_success=on_compute_success,
                plot_error=on_plot_error,
                plot_success=on_plot_success,
            )
            if scenarios[scenario] == "debug" and self.raise_exc:
                return True
            elif scenarios[scenario] is True:
                return True
            return False

        def parse_output_path(
            scenario: str,
            *,
            plot_dir: str = None,
            path_fstr: str = "{plot_dir:}/{name:}_dag_{scenario:}.pdf",
            **fstr_kwargs,
        ) -> str:
            """Prepares the output path for the DAG visualization"""
            if plot_dir is None and "plot_dir" in path_fstr:
                if self._out_path is None:
                    raise ValueError(
                        "Missing plot output path from which to extract the "
                        "`plot_dir` argument for the DAG visualization output "
                        "path! This should not have happened; make sure your "
                        "plot creator sets the _out_path attribute before "
                        "DAG visualization is invoked."
                    )
                plot_dir = os.path.dirname(self._out_path)

            p = path_fstr.format(
                plot_dir=plot_dir,
                name=self.name,
                scenario=scenario,
                **fstr_kwargs,
            )
            return os.path.expanduser(p)

        @contextlib.contextmanager
        def exception_handling(desc: str):
            """Exception handler for parts of the DAG representation routine"""
            try:
                yield

            except Exception as exc:
                msg = f"Failed {desc}!"
                _raise = raise_exc if raise_exc is not None else self.raise_exc
                if not _raise:
                    log.warning(msg)
                    log.note("Enable debug mode to show traceback.")
                    return

                raise PlotCreatorError(
                    f"{msg}\n"
                    "Inspect the chained traceback for more information "
                    "or disable debug mode to ignore this error message.\n\n"
                    f"{type(exc).__name__}: {exc}"
                ) from exc

        # .....................................................................

        when = when if when else {}
        output = output if output else {}
        export = export if export else {}
        generation = generation if generation else {}

        # Decide whether to plot
        if not should_plot(scenario, enabled=enabled, **when):
            log.debug("Not creating DAG visualization.")
            return

        log.note("Creating DAG visualization (scenario: '%s') ...", scenario)

        # Create the graph object.
        # If this fails and we are not allowed to raise, we have no other
        # option but to return None; would not make sense to continue.
        g = None
        with exception_handling("generating DAG representation"):
            g = self._dag.generate_nx_graph(**generation)

        if g is None:
            return

        # Generate the output path (for the plot)
        out_path = parse_output_path(scenario, **output)

        # Set some parameter defaults
        title_fstr = "DAG @ scenario '{}'"
        default_plot_kwargs = dict(
            annotate_kwargs=dict(
                title=title_fstr.format(scenario.replace("_", " ")),
                add_legend=True,
            )
        )
        plot_kwargs = recursive_update(
            default_plot_kwargs, copy.deepcopy(plot_kwargs)
        )

        # Export the graph object
        if export_enabled:
            from ...utils.nx import export_graph

            with exception_handling("exporting DAG representation"):
                export_graph(g, out_path=out_path, **export)

        # Plot it
        if plot_enabled:
            with exception_handling("plotting DAG representation"):
                self._dag.visualize(g=g, out_path=out_path, **plot_kwargs)

                if "error" in scenario:
                    log.caution(
                        "Created DAG visualization for scenario '%s'. "
                        "For debugging, inspecting the generated plot and the "
                        "traceback information may be helpful:\n  %s",
                        scenario,
                        out_path,
                    )

        # All done
        self._dag_vis_done_for.append(scenario)

        return g



"""networkx-related utility functions"""

import copy
import logging
import os
from typing import Any, Callable, Dict, Sequence, Union

from .._dag_utils import parse_dag_minimal_syntax as _parse_dag_minimal_syntax
from .._dag_utils import parse_dag_syntax as _parse_dag_syntax
from ..data_ops import apply_operation as _apply_operation
from ..data_ops import is_operation as _is_operation
from ..exceptions import *

log = logging.getLogger(__name__)

ATTR_MAPPER_OP_PREFIX = "attr_mapper"
"""A prefix used for registring attribute mapping data operations"""

ATTR_MAPPER_OP_PREFIX_DAG = f"{ATTR_MAPPER_OP_PREFIX}.dag"
"""A prefix used for registring attribute mapping data operations that are
specialized for use in the DAG, e.g. in
:py:meth:`dantro.dag.TransformationDAG.generate_nx_graph`.
"""

# -----------------------------------------------------------------------------


def keep_node_attributes(g: "networkx.Graph", *to_keep):
    """Iterates over the given graph object and removes all node attributes
    *but* those in ``to_keep``.

    .. note:: This function works in-place on the given graph object

    Args:
        g (networkx.Graph): The graph object with the nodes
        *to_keep: Sequence of attribute names to *keep*
    """
    for _, node_attrs in g.nodes(data=True):
        to_pop = [a for a in node_attrs if a not in to_keep]
        for a in to_pop:
            node_attrs.pop(a, None)


def keep_edge_attributes(g: "networkx.Graph", *to_keep):
    """Iterates over the given graph object and removes all edge attributes
    *but* those in ``to_keep``.

    .. note:: This function works in-place on the given graph object

    Args:
        g (networkx.Graph): The graph object with the edges
        *to_keep: Sequence of attribute names to *keep*
    """
    for _, _, edge_attrs in g.edges(data=True):
        to_pop = [a for a in edge_attrs if a not in to_keep]
        for a in to_pop:
            edge_attrs.pop(a, None)


def map_attributes(
    g: "networkx.Graph", kind: str, mappers: Dict[str, Union[str, dict]]
):
    """Maps attributes of nodes or edges (specified by ``kind``).

    Args:
        g (networkx.Graph): Graph object to map the node or edge attributes of
        kind (str): Name of a valid graph iterator, e.g. ``nodes``, ``edges``
        mappings (Dict[str, Union[str, Callable]]): The mappings dict.
            Will set node attributes that have as their value the result of a
            single data operation. The dict values can either be the name of a
            registered data operation or a dict that defines an operation and
            the corresponding arguments, supporting the
            :ref:`typical DAG syntax <dag_minimal_syntax>`.

            .. note::

                Note that the operation needs to be part of an extended
                dantro operations database. It may *not* be a
                meta-operation and can also not be a *sequence* of
                operations. The operation will always get node's or edge's
                existing ``attrs`` dict as a keyword argument. The return value
                of the operation is used as the new attribute value.
    """

    def parse_op_params(p: Union[str, dict]) -> dict:
        """Parses operation parameters using the usual DAG syntax"""
        p = _parse_dag_minimal_syntax(p, with_previous_result=False)
        return _parse_dag_syntax(**p)

    # Prepare the mappers
    mappers = mappers if mappers else {}
    mappers = {
        attr: parse_op_params(op_params)
        for attr, op_params in mappers.items()
        if op_params is not None
    }

    # Prepare the node or edge iterator and then start iterating ...
    obj_it = getattr(g, kind)(data=True)

    for obj_and_attrs in obj_it:
        attrs = obj_and_attrs[-1]

        for target_attr, op_params in mappers.items():
            try:
                attrs[target_attr] = _apply_operation(
                    op_params["operation"],
                    *op_params["args"],
                    attrs=attrs,
                    **op_params["kwargs"],
                )

            except BadOperationName as err:
                _prefix = ATTR_MAPPER_OP_PREFIX
                raise BadOperationName(
                    f"Failed mapping {kind}' attributes due to an invalid "
                    f"operation name. Use operations prefixed with {_prefix} "
                    f"for common attribute mapping tasks.\n\n{err}"
                ) from err

            except Exception as exc:
                _op_name = op_params["operation"]
                raise type(exc)(
                    f"Failed mapping {kind}' attributes:\n"
                    f"Make sure the data operation ({_op_name}) and arguments "
                    "are valid and inspect the chained traceback for more "
                    f"information.\n\nGot a {type(exc).__name__}: {exc}"
                ) from exc


def manipulate_attributes(
    g: "networkx.Graph",
    *,
    map_node_attrs: Dict[str, Union[str, Callable]] = None,
    map_edge_attrs: Dict[str, Union[str, Callable]] = None,
    keep_node_attrs: Union[bool, Sequence[str]] = True,
    keep_edge_attrs: Union[bool, Sequence[str]] = True,
):
    """Manipulates the given graph's edge and/or node attributes

    Args:
        g (networkx.Graph): The graph the node and edge attributes of which are
            to be manipulated
        map_node_attrs (Dict[str, Union[str, Callable]], optional): Sets
            the node attributes given by the keys of this dict with those at
            the value. If a callable is given, is invoked with the unpacked
            dict of node attributes as arguments and writes the return value
            to the attribute given by the key.
        map_edge_attrs (Dict[str, Union[str, Callable]], optional): Sets
            the edge attributes given by the keys of this dict with those at
            the value. If a callable is given, is invoked with the unpacked
            dict of edge attributes as arguments and writes the return value
            to the attribute given by the key.
        keep_node_attrs (Union[bool, Sequence[str]], optional): Which node
            attributes to *keep*, all others are dropped. Set to True to keep
            all existing node attributes; for all other values the
            :py:func:`~.keep_node_attributes` function is invoked.
        keep_edge_attrs (Union[bool, Sequence[str]], optional): Which edge
            attributes to *keep*, all others are dropped. Set to True to keep
            all existing edge attributes; for all other values the
            :py:func:`~.keep_edge_attributes` function is invoked.
    """
    if map_node_attrs:
        map_attributes(g, "nodes", map_node_attrs)

    if map_edge_attrs:
        map_attributes(g, "edges", map_edge_attrs)

    # Keep only certain node and/or edge attributes
    if keep_node_attrs is not True:
        keep_node_attrs = keep_node_attrs if keep_node_attrs else ()
        keep_node_attributes(g, *(keep_node_attrs if keep_node_attrs else ()))

    if keep_edge_attrs is not True:
        keep_edge_attrs = keep_edge_attrs if keep_edge_attrs else ()
        keep_edge_attributes(g, *(keep_edge_attrs if keep_edge_attrs else ()))


def export_graph(
    g: "networkx.Graph",
    *,
    out_path: str,
    manipulate_attrs: dict = None,
    **export_specs,
):
    """Takes care of exporting a networkx graph object using one or many of the
    ``nx.write_`` methods. See the
    `networkx documentation <https://networkx.org/documentation/stable/reference/readwrite/>`_
    for available output formats.

    This also allows some pre-processing or node and edge attributes using the
    :py:func:`.manipulate_attributes` function.

    Example:

    .. code-block:: yaml

        manipulate_attrs: {}

        # Export formats
        dot: true  # needs graphviz
        graphml:
          file_ext: gml
          # ... further kwargs passed to writer
        gml: false

    Args:
        g (networkx.Graph): The graph to export
        out_path (str): Path to export it to; extensions will be dropped and
            replaced by the corresponding export format. Add the ``file_ext``
            key to a export format specification to set it to some other value.
        manipulate_attrs (dict, optional): If given, is passed to
            :py:func:`~dantro.utils.nx.manipulate_attributes` to manipulate the
            node and/or edge attributes of a (copy of) the given graph ``g``.
        **export_specs: Keys need to correspond to valid ``nx.write_*``
            function names, values are passed on to the write function. There
            are two special keys ``enabled`` and ``file_ext`` that can control
            the behaviour of the respective export operation.
            Alternatively, values can be a boolean that enables or disables
            the writer.
    """
    import networkx as nx

    NX_WRITERS = {
        f[6:]: getattr(nx, f) for f in dir(nx) if f.startswith("write_")
    }
    try:
        import nx.drawing.nx_agraph

        NX_WRITERS["dot"] = nx.drawing.nx_agraph.write_dot

    except ImportError:
        pass

    if manipulate_attrs:
        # Need to work on a copy because certain attributes may be dropped
        g = copy.deepcopy(g)
        manipulate_attributes(g, **manipulate_attrs)

    # No go over the export specifications
    export_specs = copy.deepcopy(export_specs)
    for export_format, specs in export_specs.items():
        if isinstance(specs, bool):
            specs = dict(enabled=specs)

        if not specs.pop("enabled", True):
            continue

        try:
            writer = NX_WRITERS[export_format]

        except KeyError as err:
            _avail = ", ".join(NX_WRITERS)
            raise ValueError(
                f"Invalid export format '{export_format}'! "
                "No such writer available in networkx.\n"
                f"Available formats: {_avail}"
            ) from err

        log.remark("Exporting %s using %s writer ...", g, export_format)

        file_ext = specs.pop("file_ext", export_format)
        out_path = f"{os.path.splitext(out_path)[0]}.{file_ext}"
        os.makedirs(os.path.dirname(out_path), exist_ok=True)
        writer(g, out_path, **specs)


# -----------------------------------------------------------------------------
# -- Custom attribute mappers -------------------------------------------------
# -----------------------------------------------------------------------------
# These rely on ``attrs["obj"]`` containing the associated node object.


@_is_operation(f"{ATTR_MAPPER_OP_PREFIX}.copy_from_attr")
def copy_from_attr(attr_to_copy_from: str, *, attrs: dict):
    """:py:func:`Attribute mapper operation <map_attributes>` that copies an
    attribute by name.
    """
    return copy.copy(attrs[attr_to_copy_from])


@_is_operation(f"{ATTR_MAPPER_OP_PREFIX}.set_value")
def set_value(value: Any, *, attrs: dict):
    """:py:func:`Attribute mapper operation <map_attributes>` that simply sets
    a value, regardless of other attributes.
    """
    return value


# .. Specializations for use within the DAG ...................................


@_is_operation(f"{ATTR_MAPPER_OP_PREFIX_DAG}.get_operation")
def get_operation(*, attrs: dict) -> str:
    """:py:func:`Attribute mapper operation <map_attributes>` that returns the
    transformation's operation name.
    See :py:attr:`dantro.dag.Transformation.operation`.

    Used in :ref:`dag_graph_vis`.
    """
    obj = attrs["obj"]
    return getattr(obj, "operation", "")


@_is_operation(f"{ATTR_MAPPER_OP_PREFIX_DAG}.get_meta_operation")
def get_meta_operation(*, attrs: dict) -> str:
    """:py:func:`Attribute mapper operation <map_attributes>` that returns the
    transformation's meta-operation name, *if* it was added as part of a meta-
    operation. Otherwise returns an empty string. This information stems from
    the :py:attr:`dantro.dag.Transformation.context` attribute.

    Used in :ref:`dag_graph_vis`.
    """
    obj = attrs["obj"]
    return getattr(obj, "context", {}).get("meta_operation", "")


@_is_operation(f"{ATTR_MAPPER_OP_PREFIX_DAG}.format_arguments")
def format_arguments(*, attrs: dict, join_str: str = "\n") -> str:
    """:py:func:`Attribute mapper operation <map_attributes>` that formats
    node arguments in a nice and readable way.

    Used in :ref:`dag_graph_vis`.
    """
    obj = attrs["obj"]
    if not hasattr(obj, "_args") or not hasattr(obj, "_kwargs"):
        return ""

    _args = join_str.join(f"{v}" for _, v in enumerate(obj._args))
    _kwargs = join_str.join(f"{k}: {v} " for k, v in obj._kwargs.items())
    return f"args: {_args}\nkwargs: {_kwargs}"
    # TODO Improve, e.g. by hiding reference hashes


@_is_operation(f"{ATTR_MAPPER_OP_PREFIX_DAG}.get_layer")
def get_layer(*, attrs: dict) -> int:
    """:py:func:`Attribute mapper operation <map_attributes>` that returns the
    transformation's layer value.
    See :py:attr:`dantro.dag.Transformation.layer`.

    Used in :ref:`dag_graph_vis`.
    """
    obj = attrs["obj"]
    return getattr(obj, "layer", 0)


@_is_operation(f"{ATTR_MAPPER_OP_PREFIX_DAG}.get_status")
def get_status(*, attrs: dict) -> str:
    """:py:func:`Attribute mapper operation <map_attributes>` that returns the
    transformation's status value.
    See :py:attr:`dantro.dag.Transformation.status`.

    Used in :ref:`dag_graph_vis`.
    """
    obj = attrs["obj"]
    return getattr(obj, "status", "none")


@_is_operation(f"{ATTR_MAPPER_OP_PREFIX_DAG}.get_description")
def get_description(
    *,
    attrs: dict,
    join_str: str = "\n",
    to_include: list = ("operation", "meta_operation", "tag", "result"),
    abbreviate_result: int = 12,
) -> str:
    """:py:func:`Attribute mapper operation <map_attributes>` that creates a
    description string from the transformation.

    Used in :ref:`dag_graph_vis`.

    Args:
        attrs (dict): Node attributes dict
        join_str (str, optional): How to join the individual segments together
        to_include (list, optional): Which segments to include.
            Can be ``'all'`` or a sequence of keys referring to individual
            segments. Available segments:

                - ``operation``
                - ``meta_operation``
                - ``tag``
                - ``result`` (if available)
                - ``status`` (if available)

            Note that the order is also given by the order in this list.
    """
    obj = attrs["obj"]
    tag = attrs.get("tag")

    # Operation
    op = getattr(obj, "operation", "")

    # Result
    result_str = ""
    if attrs.get("has_result"):
        result = attrs.get("result")
        result_str = str(result)

        if abbreviate_result:
            # Use only the first line
            result_str = result_str.split("\n")[0]

            # If there are more characters, abbreviate to type
            if len(result_str) > abbreviate_result:
                result_str = str(type(result).__name__)

    # Status
    status = getattr(obj, "status", "none").replace("_", " ")

    # Meta-operation this node may have been part of
    meta_op = getattr(obj, "context", {}).get("meta_operation")

    # Assemble, evaluate and join descriptions
    desc_specs = dict(
        operation=dict(fstr="{}", content=op),
        meta_operation=dict(fstr="({})", content=meta_op),
        result=dict(fstr="= {}", content=result_str),
        tag=dict(fstr="— {} —", content=tag),
        status=dict(fstr="status: {}", content=status),
    )
    if to_include == "all":
        to_include = list(desc_specs.keys())

    return join_str.join(
        desc_specs[name]["fstr"].format(desc_specs[name]["content"])
        for name in to_include
        if desc_specs[name]["content"]
    )



"""A data group that represents a directory path and holds the corresponding
file paths as members"""

import logging
import pathlib
from typing import Optional, Union

from ..containers import PathContainer
from ..exceptions import ItemAccessError
from . import OrderedDataGroup, is_group

log = logging.getLogger(__name__)


@is_group
class DirectoryGroup(OrderedDataGroup):
    """A group that maps to a directory path in the file system."""

    _NEW_CONTAINER_CLS: type = PathContainer
    _ALLOWED_CONT_TYPES: tuple = ("self", PathContainer)

    def __init__(
        self,
        *args,
        name: str,
        dirpath: Union[str, pathlib.Path] = None,
        strict: bool = True,
        parent: "DirectoryGroup" = None,
        **kwargs,
    ):
        """Sets up a DirectoryGroup instance that holds the path to a certain
        directory.

        By default, this group only allows members to be
        :py:class:`~dantro.containers.path.PathContainer` or
        :py:class:`~dantro.groups.dirpath.DirectoryGroup` instances in order to
        maintain an association to the filesystem directory represented by this
        group.

        .. note::

            It is not actually checked whether the path points to a directory.

        Args:
            *args: Passed to parent init
            dirpath (Union[str, pathlib.Path], optional): A path compatible to
                :py:class:`pathlib.Path`. This path need not exist in the file
                system but it should point to a directory.
                If not given at initialization, it should be set afterwards.
            strict (bool, optional): If not True, will allow members of this
                group to be of *any* kind of dantro object.
            **kwargs: Passed to parent init
        """
        super().__init__(*args, name=name, parent=parent, **kwargs)

        if dirpath is None:
            if parent is not None and isinstance(parent, DirectoryGroup):
                dirpath = parent.fs_path.joinpath(name)
            else:
                raise TypeError(
                    f"If not supplying `dirpath` argument to {self.logstr}, "
                    "need a parent that is a DirectoryGroup and that can be "
                    "used to deduce a directory path. "
                    f"Parent was:  {parent}"
                )

        self._fs_path = pathlib.Path(dirpath)

        if not strict:
            self._ALLOWED_CONT_TYPES = None

    @property
    def fs_path(self) -> pathlib.Path:
        """Returns the filesystem path associated with this group by reading
        the corresponding data attribute (by default: ``fs_path``).

        .. note::

            The path this object represents may or may not exist.
        """
        return self._fs_path



"""This module implements the base classes of dantro, based on the abstract
classes implemented in :py:mod:`dantro.abc`.

The base classes are classes that combine features of the abstract classes.
For example, the data group gains attribute functionality by being a
combination of the :py:class:`~dantro.abc.AbstractDataGroup` and the
:py:class:`.BaseDataContainer`. In turn, the :py:class:`.BaseDataContainer`
uses the :py:class:`.BaseDataAttrs` class as an attribute and thereby extends
the :py:class:`~dantro.abc.AbstractDataContainer` class.

.. note::

    These classes are not meant to be instantiated but used as a *basis* to
    implement more specialized :py:class:`.BaseDataGroup`- or
    :py:class:`.BaseDataContainer`-derived classes.
"""

import abc
import copy
import inspect
import logging
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from .abc import (
    PATH_JOIN_CHAR,
    AbstractDataAttrs,
    AbstractDataContainer,
    AbstractDataGroup,
    AbstractDataProxy,
)
from .exceptions import ExistingDataError, ExistingGroupError, ItemAccessError
from .mixins import (
    AttrsMixin,
    BasicComparisonMixin,
    CheckDataMixin,
    CollectionMixin,
    DirectInsertionModeMixin,
    ItemAccessMixin,
    LockDataMixin,
    MappingAccessMixin,
    SizeOfMixin,
)

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class BaseDataProxy(AbstractDataProxy):
    """The base class for data proxies.

    .. note::

        This is still an abstract class and needs to be subclassed.
    """

    _tags: tuple = tuple()
    """Associated tags.

    These are empty by default and may also be overwritten in the object.
    """

    @abc.abstractmethod
    def __init__(self, obj: Any = None):
        """Initialize a proxy object for the given object."""
        if obj is not None:
            log.trace("Initialising %s for %s ...", self.classname, type(obj))
        else:
            log.trace("Initialising %s ...", self.classname)

    @property
    def tags(self) -> Tuple[str]:
        """The tags describing this proxy object"""
        return self._tags


# -----------------------------------------------------------------------------


class BaseDataAttrs(MappingAccessMixin, AbstractDataAttrs):
    """A class to store attributes that belong to a data container.

    This implements a dict-like interface and serves as default attribute
    class.

    .. note::

        Unlike the other base classes, this can already be instantiated. That
        is required as it is needed in BaseDataContainer where no previous
        subclassing or mixin is reasonable.
    """

    def __init__(self, attrs: Dict[str, Any] = None, **dc_kwargs):
        """Initialize a DataAttributes object.

        Args:
            attrs (Dict[str, Any], optional): The attributes to store
            **dc_kwargs: Further kwargs to the parent DataContainer
        """
        # Make sure it is a dict; initialize empty if empty
        attrs = dict(attrs) if attrs else {}

        super().__init__(data=attrs, **dc_kwargs)

    # .........................................................................

    def as_dict(self) -> dict:
        """Returns a shallow copy of the data attributes as a dict"""
        return {k: v for k, v in self.items()}

    # .........................................................................
    # Magic methods and iterators for convenient dict-like access

    def _format_info(self) -> str:
        """A __format__ helper function: returns info about these attributes"""
        return f"{len(self)} attribute(s)"


# -----------------------------------------------------------------------------


class BaseDataContainer(
    AttrsMixin,
    SizeOfMixin,
    BasicComparisonMixin,
    AbstractDataContainer,
):
    """The BaseDataContainer extends the abstract base class by the ability to
    hold attributes and be path-aware.
    """

    _ATTRS_CLS: type = BaseDataAttrs
    """The class to use for storing attributes"""

    def __init__(
        self,
        *,
        name: str,
        data: Any,
        attrs: Dict[str, Any] = None,
        parent: "AbstractDataGroup" = None,
    ):
        """Initialize a BaseDataContainer, which can store data and attributes.

        Args:
            name (str): The name of this data container
            data (Any): The data to store in this container
            attrs (Dict[str, Any], optional): A mapping that is stored as
                data attributes.
            parent (AbstractDataGroup, optional): If known, the parent group,
                which can be used to extract information during initialization.
                Note that linking occurs only after the container was added to
                the parent group using the
                :py:meth:`~dantro.base.BaseDataGroup.add` method. The child
                object is not responsible of linking or adding itself to the
                group.
        """
        # Initialize via parent, then additionally store attributes
        super().__init__(name=name, data=data)
        self.attrs = attrs

    def _format_info(self) -> str:
        """A __format__ helper function: returns info about the content of this
        data container.
        """
        return "{} attribute{}".format(
            len(self.attrs), "s" if len(self.attrs) != 1 else ""
        )


# -----------------------------------------------------------------------------


class BaseDataGroup(
    LockDataMixin,
    AttrsMixin,
    SizeOfMixin,
    BasicComparisonMixin,
    DirectInsertionModeMixin,
    AbstractDataGroup,
):
    """The BaseDataGroup serves as base group for all data groups.

    It implements all functionality expected of a group, which is much more
    than what is expected of a general container.
    """

    _ATTRS_CLS: type = BaseDataAttrs
    """Which class to use for storing attributes"""

    _STORAGE_CLS: type = dict
    """The mapping type that is used to store the members of this group."""

    _NEW_GROUP_CLS: type = None
    """Which class to use when creating a new group via :py:meth:`.new_group`.
    If None, the type of the current instance is used for the new group."""

    _NEW_CONTAINER_CLS: type = None
    """Which class to use for creating a new container via call to the
    :py:meth:`.new_container` method. If None, the type needs to be specified
    explicitly in the method call.
    """

    _DATA_GROUP_CLASSES: Dict[str, type] = None
    """Mapping from strings to available data group types. Used in
    string-based lookup of group types in :py:meth:`.new_group`.
    """

    _DATA_CONTAINER_CLASSES: Dict[str, type] = None
    """Mapping from strings to available data container types. Used in
    string-based lookup of container types in :py:meth:`.new_container`.
    """

    _ALLOWED_CONT_TYPES: Optional[tuple] = None
    """The types that are allowed to be stored in this group. If None, all
    types derived from the dantro base classes are allowed.
    This applies to both containers and groups that are added to this group.

    .. hint::

        To add the type of the current object, add a string entry ``self`` to
        the tuple. This will be resolved to ``type(self)`` at invocation.
    """

    _COND_TREE_MAX_LEVEL = 10
    """Condensed tree representation maximum level"""

    _COND_TREE_CONDENSE_THRESH = 10
    """Condensed tree representation threshold parameter"""

    # .........................................................................

    def __init__(
        self,
        *,
        name: str,
        containers: list = None,
        attrs=None,
        parent: "AbstractDataGroup" = None,
    ):
        """Initialize a BaseDataGroup, which can store other containers and
        attributes.

        Args:
            name (str): The name of this data container
            containers (list, optional): The containers that are to be stored
                as members of this group. If given, these are added one by one
                using the `.add` method.
            attrs (None, optional): A mapping that is stored as attributes
            parent (AbstractDataGroup, optional): If known, the parent group,
                which can be used to extract information during initialization.
                Note that linking occurs only after the group was added to the
                parent group, i.e. *after* initialization finished.
        """
        # Prepare the storage class that is used to store the members
        data = self._STORAGE_CLS()

        # Initialize via parent and store attributes
        super().__init__(name=name, data=data)
        self.attrs = attrs

        # Now add the member containers
        if containers is not None:
            self.add(*containers)

    # .........................................................................
    # Item access and manipulation

    def __getitem__(self, key: Union[str, List[str]]) -> AbstractDataContainer:
        """Looks up the given key and returns the corresponding item.

        This supports recursive *relative* lookups in two ways:

            * By supplying a path as a string that includes the path separator.
              For example, ``foo/bar/spam`` walks down the tree along the given
              path segments.
            * By directly supplying a key sequence, i.e. a list or tuple of
              key strings.

        With the last path segment, it *is* possible to access an element that
        is no longer part of the data tree; successive lookups thus need to
        use the interface of the corresponding leaf object of the data tree.

        Absolute lookups, i.e. from path ``/foo/bar``, are **not** possible!

        Lookup complexity is that of the underlying data structure: for groups
        based on dict-like storage containers, lookups happen in constant time.

        .. note::

            This method aims to replicate the behavior of POSIX paths.

            Thus, it can also be used to access the element itself or the
            parent element: Use ``.`` to refer to this object and ``..`` to
            access this object's ``parent``.

        Args:
            key (Union[str, List[str]]): The name of the object to retrieve or
                a path via which it can be found in the data tree.

        Returns:
            AbstractDataContainer: The object at ``key``, which concurs to the
                dantro tree interface.

        Raises:
            ItemAccessError: If no object could be found at the given ``key``
                or if an absolute lookup, starting with ``/``, was attempted.
        """
        if isinstance(key, str):
            key_seq = key.split(PATH_JOIN_CHAR)
        else:
            # Assume it is list-like ... that's all we need to assume here.
            key_seq = key

        # Do not allow absolute lookups or empty arguments
        if not key_seq or (key_seq and not key_seq[0]):
            _key = PATH_JOIN_CHAR.join(key_seq)
            raise ItemAccessError(
                self,
                key=_key,
                show_hints=False,
                suffix=(
                    "Can only do relative lookups! Remove the leading '/' "
                    "from the given path or make sure that the given key "
                    f"sequence ({key_seq}) does not start with an element "
                    "that evaluates to False."
                ),
            )

        # Remove any empty elements to allow paths like foo////bar
        key_seq = [seg for seg in key_seq if seg]

        # Now can be sure that there is at least one segment in the path
        # Have three cases now ...
        # ... next item is this item
        if key_seq[0] == ".":
            item = self

        # ... next item is the parent item
        elif key_seq[0] == "..":
            if self.parent is None:
                raise ItemAccessError(
                    self,
                    key="..",
                    show_hints=False,
                    suffix="No parent associated.",
                )
            item = self.parent

        # ... next item is a downstream item
        else:
            try:
                item = self._data[key_seq[0]]

            except (KeyError, IndexError) as err:
                _key = PATH_JOIN_CHAR.join(key_seq)
                raise ItemAccessError(self, key=_key) from err

        # If there was only one key, this is the end of the recursion.
        if len(key_seq) == 1:
            return item

        # Otherwise, we have to recursively continue with the key lookup ...
        # NOTE There deliberately is no error handling here. Further errors
        #      should be handled by the *next* item, because *so far*, all the
        #      item access was successful.
        return item[key_seq[1:]]

    def __setitem__(
        self, key: Union[str, List[str]], val: BaseDataContainer
    ) -> None:
        """This method is used to allow access to the content of containers of
        this group. For adding an element to this group, use the `add` method!

        Args:
            key (Union[str, List[str]]): The key to which to set the value.
                If this is a path, will recurse down to the lowest level.
                Note that all intermediate keys need to be present.
            val (BaseDataContainer): The value to set

        Returns:
            None

        Raises:
            ValueError: If trying to add an element to this group, which should
                be done via the `add` method.

        """
        if not isinstance(key, list):
            key = key.split(PATH_JOIN_CHAR)

        # Depending on length of the key sequence, start recursion or not
        if len(key) > 1:
            self._data[key[0]][key[1:]] = val
            return

        # else: end of recursion, i.e. the path led to an item of this group
        # This operation is not allowed, as the add method should be used
        # That method takes care that the name this element is registered with
        # is equal to that of the registered object
        raise ValueError(
            f"{self.logstr} cannot carry out __setitem__ operation for the "
            f"given key '{key}'. Note that to add a group or container to "
            "the group, the `add` method should be used."
        )

    def __delitem__(self, key: str) -> None:
        """Deletes an item from the group"""
        if not isinstance(key, list):
            # Assuming this is a string ...
            key = key.split(PATH_JOIN_CHAR)

        # Can be sure that this is a list now
        # If there is more than one entry, need to call this recursively
        if len(key) > 1:
            # Continue recursion
            del self._data[key[0]][key[1:]]
            return

        # else: end of recursion: delete and unlink this container
        # ... if it is not locked
        self.raise_if_locked()

        cont = self._data[key[0]]
        del self._data[key[0]]

        self._unlink_child(cont)

    def add(self, *conts, overwrite: bool = False):
        """Add the given containers to this group."""
        for cont in conts:
            self._add_container(cont, overwrite=overwrite)

        log.trace("Added %d container(s) to %s.", len(conts), self.logstr)

    def _add_container(self, cont, *, overwrite: bool):
        """Private helper method to add a container to this group."""
        # Data may not be locked
        self.raise_if_locked()

        # Check the allowed types
        if self._ALLOWED_CONT_TYPES is None and not isinstance(
            cont, (BaseDataGroup, BaseDataContainer)
        ):
            raise TypeError(
                "Can only add BaseDataGroup- or BaseDataContainer-derived"
                f" objects to {self.logstr}, got {type(cont)}!"
            )

        elif self._ALLOWED_CONT_TYPES is not None:
            allowed_types = [
                t if t != "self" else type(self)
                for t in self._ALLOWED_CONT_TYPES
            ]
            if not isinstance(cont, tuple(allowed_types)):
                raise TypeError(
                    "Can only add objects derived from the following "
                    f"classes: {allowed_types}. Got: {type(cont)}"
                )

        # else: is of correct type
        # Check if one like this already exists
        old_cont = None
        if cont.name in self:
            if not overwrite:
                raise ExistingDataError(
                    f"{self.logstr} already has a member with "
                    f"name '{cont.name}', cannot add {cont}."
                )
            log.debug(
                "A member '%s' of %s already exists and will be "
                "overwritten ...",
                cont.name,
                self.logstr,
            )
            old_cont = self[cont.name]

        # Allow for subclasses to perform further custom checks on the
        # container object before adding it
        self._check_cont(cont)

        # Write to data, assuring that the name matches that of the container
        self._add_container_to_data(cont)
        self._add_container_callback(cont)

        # Re-link the containers
        self._link_child(new_child=cont, old_child=old_cont)

    def _check_cont(self, cont) -> None:
        """Can be used by a subclass to check a container before adding it to
        this group. Is called by _add_container before checking whether the
        object exists or not.

        This is not expected to return, but can raise errors, if something
        did not work out as expected.

        Args:
            cont: The container to check
        """
        pass

    def _add_container_to_data(self, cont: AbstractDataContainer) -> None:
        """Performs the operation of adding the container to the _data. This
        can be used by subclasses to make more elaborate things while adding
        data, e.g. specify ordering ...

        NOTE This method should NEVER be called on its own, but only via the
             _add_container method, which takes care of properly linking the
             container that is to be added.

        NOTE After adding, the container need be reachable under its .name!

        Args:
            cont: The container to add
        """
        # Just add it via _data.__setitem__, using the container's name
        self._data[cont.name] = cont

    def _add_container_callback(self, cont) -> None:
        """Called after a container was added."""
        pass

    def new_container(
        self,
        path: Union[str, List[str]],
        *,
        Cls: Union[type, str] = None,
        GroupCls: Union[type, str] = None,
        _target_is_group: bool = False,
        **kwargs,
    ) -> "BaseDataContainer":
        """Creates a new container of type ``Cls`` and adds it at the given
        path relative to this group.

        If needed, intermediate groups are automatically created.

        Args:
            path (Union[str, List[str]]): Where to add the container.
            Cls (Union[type, str], optional): The type of the target container
                (or group) that is to be added.
                If None, will use the type set in ``_NEW_CONTAINER_CLS`` class
                variable. If a string is given, the type is looked up in the
                container type registry.
            GroupCls (Union[type, str], optional): Like ``Cls`` but used for
                intermediate group types only.
            _target_is_group (bool, optional): Internally used variable.
                If True, will look up the ``Cls`` type via
                :py:meth:`._determine_group_type` instead of
                :py:meth:`._determine_container_type`.
            **kwargs: passed on to ``Cls.__init__``

        Returns:
            BaseDataContainer:
                The created container of type ``Cls``
        """
        # Make sure the path is a list and of valid content
        if isinstance(path, str):
            path = path.split(PATH_JOIN_CHAR)
        path = list(path)

        if not path or not path[0]:
            raise ValueError(f"`path` argument may not be empty! Got: {path}")

        # Check whether recursion ends here, i.e.: the path ends here
        if len(path) == 1:
            if not _target_is_group:
                Cls = self._determine_container_type(Cls)
            else:
                Cls = self._determine_group_type(Cls)

            # Create the container and add it
            # Let the container know about its parent here, in case it needs to
            # do setup steps depending on that. However, linking etc will not
            # be done there but in the `add` call after that.
            cont = Cls(name=path[0], parent=self, **kwargs)
            self.add(cont)
            return cont

        # -- Recursive branch
        # Need to split off the front section and continue
        grp_name, new_path = path[0], path[1:]

        # Retrieve the group, creating it if it does not exist
        if grp_name not in self:
            # Use the group class for intermediate paths.
            # grp_name being a single string will lead into the non-recursive
            # branch of this method again.
            grp = self.new_group(grp_name, Cls=GroupCls)
        else:
            grp = self[grp_name]

            if not isinstance(grp, BaseDataGroup):
                raise ExistingDataError(
                    f"Tried to create a new group '{grp_name}' in "
                    f"{self.logstr}, but a container was already stored "
                    "at that path."
                )

        # Can now create the container, potentially recursively creating more
        # intermediate groups along the path ...
        return grp.new_container(
            new_path,
            Cls=Cls,
            GroupCls=GroupCls,
            _target_is_group=_target_is_group,
            **kwargs,
        )

    def new_group(
        self,
        path: Union[str, List[str]],
        *,
        Cls: Union[type, str] = None,
        GroupCls: Union[type, str] = None,
        **kwargs,
    ) -> "BaseDataGroup":
        """Creates a new group at the given path.

        Args:
            path (Union[str, List[str]]): The path to create the group at.
                If necessary, intermediate paths will be created.
            Cls (Union[type, str], optional): If given, use this type to
                create the target group. If not given, uses the class
                specified in the ``_NEW_GROUP_CLS`` class variable or (if a
                string) the one from the group type registry.

                .. note::

                    This argument is evaluated at each segment of the ``path``
                    by the corresponding object in the tree. Subsequently, the
                    types need to be available at the desired

            GroupCls (Union[type, str], optional): Like ``Cls``, but this
                applies only to the creation of intermediate groups.
            **kwargs: Passed on to ``Cls.__init__``

        Returns:
            BaseDataGroup:
                The created group of type ``Cls``
        """
        # Use container method to create the entry. Recursion happens there.
        return self.new_container(
            path, Cls=Cls, GroupCls=GroupCls, _target_is_group=True, **kwargs
        )

    def recursive_update(self, other, *, overwrite: bool = True):
        """Recursively updates the contents of this data group with the entries
        of the given data group

        .. note::

            This will create *shallow* copies of those elements in ``other``
            that are added to this object.

        Args:
            other (BaseDataGroup): The group to update with
            overwrite (bool, optional): Whether to overwrite already existing
                object. If False, a conflict will lead to an error being
                raised and the update being stopped.

        Raises:
            TypeError: If ``other`` was of invalid type
        """

        if not isinstance(other, BaseDataGroup):
            raise TypeError(
                f"Can only update {self.logstr} with objects of classes that "
                f"are derived from BaseDataGroup. Got: {type(other)}"
            )

        # Loop over the given DataGroup
        for name, obj in other.items():
            # Distinguish between the case where it is another group and where
            # it is a container
            if isinstance(obj, BaseDataGroup):
                # Already a group -> if a group with the same name is already
                # present, continue recursion. If not, just create an entry
                # and add it to this group
                if name in self:
                    # Continue recursion
                    self[name].recursive_update(obj, overwrite=overwrite)
                else:
                    # Can add the object, but need to detach the parent first
                    # and thus need to work on a copy
                    obj = copy.copy(obj)
                    obj.parent = None
                    self.add(obj, overwrite=overwrite)

            else:
                # Not a group; add a shallow copy
                obj = copy.copy(obj)
                obj.parent = None
                self.add(obj, overwrite=overwrite)

        log.debug("Finished recursive update of %s.", self.logstr)

    def clear(self):
        """Clears all containers from this group.

        This is done by unlinking all children and then overwriting ``_data``
        with an empty ``_STORAGE_CLS`` object.
        """
        for child in self.values():
            self._unlink_child(child)
        self._data = self._STORAGE_CLS()
        log.debug("%s cleared.", self.logstr)

    # .........................................................................

    def _determine_container_type(self, Cls: Union[type, str]) -> type:
        """Helper function to determine the type to use for a new container.

        Args:
            Cls (Union[type, str]): If None, uses the ``_NEW_CONTAINER_CLS``
                class variable. If a string, tries to extract it from the class
                variable ``_DATA_CONTAINER_CLASSES`` dict.
                Otherwise, assumes this is already a type.

        Returns:
            type:
                The container class to use

        Raises:
            ValueError: If the string class name was not registered
            AttributeError: If no default class variable was set
        """
        default = getattr(self, "_NEW_CONTAINER_CLS", None)

        Cls = self._determine_type(
            Cls,
            default=default,
            registry=getattr(self, "_DATA_CONTAINER_CLASSES", None),
        )
        if inspect.isclass(Cls) and not issubclass(
            Cls, (BaseDataContainer, BaseDataGroup)
        ):
            raise TypeError(
                "Expected a subclass of BaseDataContainer or BaseDataGroup, "
                f"got {Cls}."
            )

        return Cls

    def _determine_group_type(self, Cls: Union[type, str]) -> type:
        """Helper function to determine the type to use for a new group.

        Args:
            Cls (Union[type, str]): If None, uses the ``_NEW_GROUP_CLS`` class
                variable. If that one is not set, uses ``type(self)``.
                If a string, tries to extract it from the class variable
                ``_DATA_GROUP_CLASSES`` dict.
                Otherwise, assumes ``Cls`` is already a type.

        Returns:
            type:
                The group class to use

        Raises:
            ValueError: If the string class name was not registered
            AttributeError: If no default class variable was set
        """
        default = getattr(self, "_NEW_GROUP_CLS", None)
        if default is None:
            default = type(self)

        Cls = self._determine_type(
            Cls,
            default=default,
            registry=getattr(self, "_DATA_GROUP_CLASSES", None),
        )

        # Need to catch the case where a non-group class was given
        if inspect.isclass(Cls) and not issubclass(Cls, BaseDataGroup):
            raise TypeError(
                f"Expected a subclass of BaseDataGroup, got {Cls}."
            )

        return Cls

    def _determine_type(
        self, T: Union[type, str], *, default: type, registry: Dict[str, type]
    ) -> type:
        """Helper function to determine a type by name, falling back to a
        default type or looking it up from a dict-like registry if it is a
        string."""
        if T is None:
            if not default:
                raise AttributeError(
                    f"Missing default container type for {self.logstr}!"
                )
            return default

        if not isinstance(T, type):
            cls_name = T

            if registry is None:
                raise AttributeError(
                    f"No type registry available for {self.logstr}! Cannot "
                    f"look up type by name '{cls_name}'."
                )

            elif cls_name not in registry:
                _avail = "\n".join(f"  {k}: {v}" for k, v in registry.items())
                raise ValueError(
                    f"The given class name '{cls_name}' was not registered "
                    f"with this {self.classname}!\nAvailable types:\n{_avail}"
                )

            # everything ok, retrieve the class type
            return registry[cls_name]

        # else: assume it is already a type and just return the given argument
        return T

    # .........................................................................
    # Linking: For correct child-parent linking, some helper methods

    def _link_child(
        self,
        *,
        new_child: BaseDataContainer,
        old_child: BaseDataContainer = None,
    ):
        """Links the new_child to this class, unlinking the old one.

        This method should be called from any method that changes which items
        are associated with this group.
        """
        # Check that it was already associated with the group
        if new_child not in self:
            raise ValueError(
                f"{new_child.logstr} needs to be a child of {self.logstr} "
                "_before_ it can be linked. Use the add method to add the "
                "child to the group."
            )
        new_child.parent = self

        if old_child is not None:
            self._unlink_child(old_child)

    def _unlink_child(self, child: BaseDataContainer):
        """Unlink a child from this class.

        This method should be called from any method that removes an item from
        this group, be it through deletion or through
        """
        if child.parent is not self:
            raise ValueError(
                f"{child.logstr} was not linked to {self.logstr}. "
                "Refuse to unlink."
            )
        child.parent = None

    # .........................................................................
    # Information

    def __len__(self) -> int:
        """The number of members in this group."""
        return len(self._data)

    def __contains__(self, cont: Union[str, AbstractDataContainer]) -> bool:
        """Whether the given container is in this group or not.

        If this is a data tree object, it will be checked whether this
        *specific* instance is part of the group, using ``is``-comparison.

        Otherwise, assumes that ``cont`` is a valid argument to the
        :py:meth:`~dantro.base.BaseDataGroup.__getitem__` method (a key or key
        sequence) and tries to access the item at that path, returning ``True``
        if this succeeds and ``False`` if not.

        Lookup complexity is that of item lookup (scalar) for both name and
        object lookup.

        Args:
            cont (Union[str, AbstractDataContainer]): The name of the
                container, a path, or an object to check via identity
                comparison.

        Returns:
            bool: Whether the given container object is part of this group or
                whether the given path is accessible from this group.
        """
        if isinstance(cont, AbstractDataContainer):
            # Case: look for the specific object instance
            # Don't iterate, as this scales badly; instead retrieve the name
            # and do the identiy lookup afterwards
            _cont = self.get(cont.name)
            if _cont is None:
                return False
            return _cont is cont

        # Otherwise: look for an object reachable at this path ...
        try:
            self[cont]
        except Exception:
            return False
        return True

    def _ipython_key_completions_(self) -> List[str]:
        """For ipython integration, return a list of available keys"""
        return list(self.keys())

    # .........................................................................
    # Iteration

    def __iter__(self):
        """Returns an iterator over the OrderedDict"""
        return iter(self._data)

    def keys(self):
        """Returns an iterator over the container names in this group."""
        return self._data.keys()

    def values(self):
        """Returns an iterator over the containers in this group."""
        return self._data.values()

    def items(self):
        """Returns an iterator over the (name, data container) tuple of this
        group."""
        return self._data.items()

    def get(self, key, default=None):
        """Return the container at `key`, or `default` if container with name
        `key` is not available."""
        return self._data.get(key, default)

    def setdefault(self, key, default=None):
        """This method is not supported for a data group"""
        raise NotImplementedError(
            f"setdefault is not supported by {self.classname}! Use the "
            "`add` method or `new_group` and `new_container` to add elements."
        )

    # .........................................................................
    # Formatting

    @property
    def tree(self) -> str:
        """Returns the default (full) tree representation of this group"""
        return self._tree_repr()

    @property
    def tree_condensed(self) -> str:
        """Returns the condensed tree representation of this group. Uses the
        ``_COND_TREE_*`` prefixed class attributes as parameters.
        """
        return self._tree_repr(
            max_level=self._COND_TREE_MAX_LEVEL,
            condense_thresh=self._COND_TREE_CONDENSE_THRESH,
        )

    def _format_info(self) -> str:
        """A __format__ helper function: returns an info string that is used
        to characterize this object. Does NOT include name and classname!
        """
        return "{} member{}, {} attribute{}".format(
            len(self),
            "s" if len(self) != 1 else "",
            len(self.attrs),
            "s" if len(self.attrs) != 1 else "",
        )

    def _format_tree(self) -> str:
        """Returns the default tree representation of this group by invoking
        the .tree property
        """
        return self.tree

    def _format_tree_condensed(self) -> str:
        """Returns the default tree representation of this group by invoking
        the .tree property
        """
        return self.tree_condensed

    def _tree_repr(
        self,
        *,
        level: int = 0,
        max_level: int = None,
        info_fstr="<{:cls_name,info}>",
        info_ratio: float = 0.6,
        condense_thresh: Union[int, Callable[[int, int], int]] = None,
        total_item_count: int = 0,
    ) -> Union[str, List[str]]:
        """Recursively creates a multi-line string tree representation of this
        group. This is used by, e.g., the _format_tree method.

        Args:
            level (int, optional): The depth within the tree
            max_level (int, optional): The maximum depth within the tree;
                recursion is not continued beyond this level.
            info_fstr (str, optional): The format string for the info string
            info_ratio (float, optional): The width ratio of the whole line
                width that the info string takes
            condense_thresh (Union[int, Callable[[int, int], int]], optional):
                If given, this specifies the threshold beyond which the tree
                view for the current element becomes condensed by hiding the
                output for some elements.
                The minimum value for this is 3, indicating that there should
                be at most 3 lines be generated from this level (excluding the
                lines coming from recursion), i.e.: two elements and one line
                for indicating how many values are hidden.
                If a smaller value is given, this is silently brought up to 3.
                Half of the elements are taken from the beginning of the
                item iteration, the other half from the end.
                If given as integer, that number is used.
                If a callable is given, the callable will be invoked with the
                current level, number of elements to be added at this level,
                and the current total item count along this recursion branch.
                The callable should then return the number of lines to be
                shown for the current element.
            total_item_count (int, optional): The total number of items
                already created in this recursive tree representation call.
                Passed on between recursive calls.

        Returns:
            Union[str, List[str]]: The (multi-line) tree representation of
                this group. If this method was invoked with ``level == 0``, a
                string will be returned; otherwise, a list of strings will be
                returned.
        """

        def get_offset_str(level: int) -> str:
            """Returns an offst string, depending on level"""
            return "   " * level

        def truncate(s: str, *, max_length: int, suffix: str = "…") -> str:
            """Truncates the given string to the desired length"""
            return s if len(s) <= max_length else s[: max_length - 1] + suffix

        # Offset
        offset = get_offset_str(level)

        # Mark symbols
        first_mark = r" └┬"
        base_mark = r"  ├"
        last_mark = r"  └"
        only_mark = r" └─"

        # Evaluate the condensation threshold, i.e. the maximum number of lines
        # to allow originating from this object (excluding recursion)
        num_items = len(self)
        total_item_count += num_items
        num_skipped = 0

        if callable(condense_thresh):
            max_lines = condense_thresh(
                level=level,
                num_items=num_items,
                total_item_count=total_item_count,
            )
        else:
            max_lines = condense_thresh

        if max_lines is not None:
            # Additional check for lower bound; makes visualization much easier
            max_lines = max(3, int(max_lines))

            # If there are too few items, the variable is set to None to
            # indicate regular behavior.
            if num_items - max_lines < 1:
                max_lines = None

        # Calculations that make the output line fit into one terminal line
        from .tools import TERMINAL_INFO, update_terminal_info

        update_terminal_info()

        num_cols = TERMINAL_INFO["columns"]
        info_width = int(num_cols * info_ratio)
        name_width = (num_cols - info_width) - (len(offset) + 3 + 1 + 2)

        def get_mark(n: int, *, max_n: int) -> str:
            """Returns the mark symbol depending on the iteration number.

            NOTE This uses variables from the outer scope!
            """
            if n == 0:
                if max_n == 0:
                    return only_mark
                return first_mark
            elif n == max_n:
                return last_mark
            return base_mark

        # The format string that's used to compose the whole output line
        fstr = "{offset:}{mark:>3s} {name:<{name_width}s}  {info:}"

        # Create the list to gather the lines in; add a description on level 0
        lines = []
        if level == 0:
            lines.append("")
            lines.append(f"Tree of {self:logstr,info}")

        # Go over the entries on this level and format the lines
        for n, (key, obj) in enumerate(self.items()):
            # Determine whether to show this line of the tree or not. The lines
            # in the middle of the iteration are not shown.
            # If it is not shown, the first line that is then to be shown also
            # adds a line that indicates how many items were skipped.
            if max_lines is not None:
                if max_lines // 2 <= n < (num_items - (max_lines - 1) // 2):
                    num_skipped += 1
                    continue

                elif n == (num_items - (max_lines - 1) // 2):
                    # Add the indicator line
                    lines.append(
                        fstr.format(
                            offset=offset,
                            mark=base_mark,
                            name_width=name_width,
                            name="...",
                            info=f"... ({num_skipped:d} more) ...",
                        )
                    )

            # Get the mark, and key and info strings (truncating if necessary)
            try:
                info_str = info_fstr.format(obj)
            except Exception as exc:
                raise ValueError(
                    f"Failed formatting info string '{info_fstr}' for "
                    f"{type(obj)} with value:\n{obj}\nThis should not have "
                    "happened! Is there a non-dantro object included as a "
                    f"direct part of the tree? Parent object: {self.logstr} "
                    f"@ {self.path}."
                ) from exc

            name = truncate(key, max_length=name_width)
            info = truncate(info_str, max_length=info_width)
            mark = get_mark(n, max_n=num_items - 1)

            # Format the line and add to list of lines
            lines.append(
                fstr.format(
                    offset=offset,
                    mark=mark,
                    name_width=name_width,
                    name=name,
                    info=info,
                )
            )

            # If it was a group and it is not empty...
            if isinstance(obj, BaseDataGroup) and len(obj) > 0:
                # ... and maximum recursion depth is not reached:
                if max_level is None or level < max_level:
                    # Continue recursion
                    lines += obj._tree_repr(
                        level=level + 1,
                        max_level=max_level,
                        info_fstr=info_fstr,
                        info_ratio=info_ratio,
                        condense_thresh=condense_thresh,
                        total_item_count=total_item_count,
                    )

                else:
                    # Only indicate that it _would_ continue here, but do not
                    # actually continue with the recursion.
                    lines.append(
                        fstr.format(
                            offset=get_offset_str(level + 1),
                            mark=only_mark,
                            name_width=3,
                            name="...",
                            info="",
                        )
                    )

        # Done, depending on whether this is within the recursion or not,
        # return as list of lines or as combined multi-line string
        if level > 0:
            return lines
        return "\n".join(lines) + "\n"



"""Tools for module importing, e.g. lazy imports."""

import copy
import importlib
import importlib.util
import logging
import os
import sys
from types import ModuleType
from typing import Any, Callable, Sequence, Tuple, Union

log = logging.getLogger(__name__)

# -- Context managers ---------------------------------------------------------


class added_sys_path:
    """A :py:data:`sys.path` context manager temporarily adding a path and
    removing it again upon exiting.
    If the given path already exists in :py:data`sys.path`, it is neither added
    nor removed and :py:data`sys.path` remains unchanged.

    .. todo::

        Expand to allow multiple paths being added
    """

    def __init__(self, path: str):
        """Initialize the context manager.

        Args:
            path (str): The path to add to :py:data:`sys.path`.
        """
        self.path = path
        self.path_already_exists = self.path in sys.path

    def __enter__(self):
        if not self.path_already_exists:
            log.debug("Temporarily adding '%s' to sys.path ...")
            sys.path.insert(0, self.path)

    def __exit__(self, *_):
        if not self.path_already_exists:
            log.debug("Removing temporarily added path from sys.path ...")
            sys.path.remove(self.path)


class temporary_sys_modules:
    """A context manager for the :py:data:`sys.modules` cache, ensuring that it
    is in the same state after exiting as it was before entering the context.

    .. note::

        This works solely on module *names*, not on the module objects! If a
        module object itself is overwritten, this context manager is not
        able to discern that as long as the key does not change.
    """

    def __init__(self, *, reset_only_on_fail: bool = False):
        """Set up the context manager for a temporary :py:data:`sys.modules`
        cache.

        Args:
            reset_only_on_fail (bool, optional): If True, will reset the cache
                only in case the context is exited with an exception.
        """
        self._modules = {ms: mod for ms, mod in sys.modules.items()}
        self.reset_only_on_fail = reset_only_on_fail

    def __enter__(self):
        pass

    def __exit__(self, exc_type: type, *_):
        if self.reset_only_on_fail and exc_type is None:
            return

        elif sys.modules == self._modules:
            return

        # else: Reset module cache
        # NOTE Cannot simply assign here, but need to insert and delete,
        #      otherwise module objects may become disassociated
        to_remove = [ms for ms in sys.modules if ms not in self._modules]
        to_add_back = [ms for ms in self._modules if ms not in sys.modules]

        for ms in to_remove:
            del sys.modules[ms]
        for ms in to_add_back:
            sys.modules[ms] = self._modules[ms]

        log.debug(
            "Reset sys.modules cache to state before entering context manager."
        )


# -- Various import functions -------------------------------------------------


def get_from_module(mod: ModuleType, *, name: str):
    """Retrieves an attribute from a module, if necessary traversing along the
    module string.

    Args:
        mod (ModuleType): Module to start looking at
        name (str): The ``.``-separated module string leading to the desired
            object.
    """
    obj = mod

    for attr_name in name.split("."):
        try:
            obj = getattr(obj, attr_name)

        except AttributeError as err:
            raise AttributeError(
                f"Failed to retrieve attribute or attribute sequence '{name}' "
                f"from module '{mod.__name__}'! Intermediate "
                f"{type(obj).__name__} {obj} has no attribute '{attr_name}'!"
            ) from err

    return obj


def import_module_or_object(
    module: str = None, name: str = None, *, package: str = "dantro"
) -> Any:
    """Imports a module or an object using the specified module string and the
    object name. Uses :py:func:`importlib.import_module` to retrieve the module
    and then uses :py:func:`~dantro._import_tools.get_from_module` for getting
    the ``name`` from that module (if given).

    Args:
        module (str, optional): A module string, e.g. ``numpy.random``. If
            this is not given, it will import from the :py:mod`builtins`
            module. If this is a relative module string, will resolve starting
            from ``package``.
        name (str, optional): The name of the object to retrieve from the
            chosen module and return. This may also be a dot-separated sequence
            of attribute names which can be used to traverse along attributes,
            which uses :py:func:`~dantro._import_tools.get_from_module`.
        package (str, optional): Where to import from if ``module`` was a
            relative module string, e.g. ``.data_mngr``, which would lead to
            resolving the module from ``<package><module>``.

    Returns:
        Any: The chosen module or object, i.e. the object found at
            ``<module>.<name>``

    Raises:
        AttributeError: In cases where part of the ``name`` argument could not
            be resolved due to a bad attribute name.
    """
    module = module if module else "builtins"
    mod = (
        importlib.import_module(module, package=package)
        if module
        else __builtin__
    )

    if not name:
        return mod
    return get_from_module(mod, name=name)


def import_name(modstr: str):
    """Given a module string, import a name, treating the last segment of the
    module string as the name.

    .. note::

        If the last segment of ``modstr`` is *not* the name, use
        :py:func:`~dantro._import_tools.import_module_or_object` instead of
        this function.

    Args:
        modstr (str): A module string, e.g. ``numpy.random.randint``, where
            ``randint`` will be the name to import.
    """
    ms = modstr.split(".")
    return import_module_or_object(module=".".join(ms[:-1]), name=ms[-1])


def import_module_from_path(
    *, mod_path: str, mod_str: str, debug: bool = True
) -> Union[None, ModuleType]:
    """Helper function to import a module that is importable only when adding
    the module's parent directory to :py:data:`sys.path`.

    .. note::

        The ``mod_path`` directory needs to contain an ``__init__.py`` file.
        If that is not the case, you cannot use this function, because the
        directory does not represent a valid Python module.

        Alternatively, a single file can be imported as a module using
        :py:func:`~dantro._import_tools.import_module_from_file`.

    Args:
        mod_path (str): Path to the module's root *directory*, ``~`` expanded
        mod_str (str): Name under which the module can be imported with
            ``mod_path`` being in :py:data:`sys.path`. This is also used to
            add the module to the :py:data:`sys.modules` cache.
        debug (bool, optional): Whether to raise exceptions if import failed

    Returns:
        Union[None, ModuleType]: The imported module or None, if importing
            failed and ``debug`` evaluated to False.

    Raises:
        ImportError: If ``debug`` is set and import failed for whatever reason
        FileNotFoundError: If ``mod_path`` did not point to an existing
            *directory*
    """
    mod_path = os.path.expanduser(mod_path)

    if not os.path.isdir(mod_path):
        raise FileNotFoundError(
            "The `mod_path` argument to import a module from a path should be "
            f"the path to an existing directory! Given path:  {mod_path}"
        )

    # Need the parent directory in the path, because the import is only
    # possible from there. This, in turn, depends on the depth of the module
    # string, so the parent directory should be chosen accordingly.
    mod_parent_dir = mod_path
    for _ in mod_str.split("."):
        mod_parent_dir = os.path.dirname(mod_parent_dir)

    try:
        # Use the temporary environments to prevent that a *failed* import ends
        # up generating an erroneous sys.modules cache entry.
        _add_sys_path = added_sys_path(mod_parent_dir)
        _tmp_sys_modules = temporary_sys_modules(reset_only_on_fail=True)
        with _add_sys_path, _tmp_sys_modules:
            mod = importlib.import_module(mod_str)

    except Exception as exc:
        if debug:
            raise ImportError(
                f"Failed importing module '{mod_str}'!\n"
                f"Make sure that {mod_path}/__init__.py can be loaded without "
                "errors (with its parent directory being part of sys.path) "
                "and that the `mod_str` argument is correct. "
                "To debug, inspect the chained traceback."
            ) from exc

        log.debug(
            "Importing module '%s' from %s failed: %s", mod_str, mod_path, exc
        )
        return

    else:
        log.debug("Successfully imported module '%s'.", mod_str)

    return mod


def import_module_from_file(
    mod_file: str,
    *,
    base_dir: str = None,
    mod_name_fstr: str = "from_file.{filename:}",
) -> ModuleType:
    """Returns the module corresponding to the file at the given ``mod_file``.

    This uses :py:func:`importlib.util.spec_from_file_location` and
    :py:func:`importlib.util.module_from_spec` to construct a module from the
    given file, regardless of whether there is a ``__init__.py`` file beside
    the file or not.

    Args:
        mod_file (str): The path to a python module *file* to load as a module
        base_dir (str, optional): If given, uses this to resolve relative
            ``mod_file`` paths.
        mod_name_fstr (str): How to name the module. Should be a format string
            that is supplied with the ``filename`` argument.

    Returns:
        ModuleType: The imported module

    Raises:
        ValueError: If ``mod_file`` was a relative path but no ``base_dir`` was
            given.
    """
    mod_file = os.path.expanduser(mod_file)

    # Evaluate relative paths
    if not os.path.isabs(mod_file):
        if not base_dir:
            raise ValueError(
                f"Cannot import from a relative `mod_file` path ({mod_file}) "
                "if no `base_dir` argument was given!"
            )

        mod_file = os.path.join(base_dir, mod_file)

    # Extract a name from the path to use as module name, needed for building
    # a module specification.
    fname, _ = os.path.splitext(os.path.basename(mod_file))
    mod_name = mod_name_fstr.format(filename=fname)

    # Create a module specification and, from that, import the module
    spec = importlib.util.spec_from_file_location(mod_name, mod_file)
    mod = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(mod)

    return mod


# -- LazyLoader ---------------------------------------------------------------


class LazyLoader:
    """Delays import until the module's attributes are accessed.

    This is inspired by an implementation by Dboy Liao, see
    `here <https://levelup.gitconnected.com/python-trick-lazy-module-loading-df9b9dc111af>`_.

    It extends on it by allowing a ``depth`` until which loading will be lazy.
    """

    def __init__(self, mod_name: str, *, _depth: int = 0):
        """Initialize a placeholder for a module.

        .. warning::

            Values of ``_depth > 0`` may lead to unexpected behaviour of the
            root module, i.e. this object, because attribute calls do not
            yield an actual object. Only use this in scenarios where you are
            in full control over the attribute calls.

            We furthermore suggest to not make the LazyLoader instance publicly
            available in such cases.

        Args:
            mod_name (str): The module name to lazy-load upon attribute call.
            _depth (int, optional): With a depth larger than zero, attribute
                calls are not leading to an import *yet*, but to the creation
                of another LazyLoader instance (with depth reduced by one).
                Note the warning above regarding usage.
        """
        self.mod_name = mod_name
        self._mod = None
        self._depth = _depth
        log.debug("LazyLoader set up for '%s'.", self.mod_name)

    def __getattr__(self, name: str):
        # If depth is zero, do the import
        if self._depth <= 0:
            if self._mod is None:
                self._mod = self.resolve()
            return getattr(self._mod, name)

        # else: create a new lazy loader for the combined module string
        return LazyLoader(f"{self.mod_name}.{name}", _depth=self._depth - 1)

    def resolve(self):
        log.debug("Now lazy-loading '%s' ...", self.mod_name)

        modstr_parts = self.mod_name.split(".")
        return import_module_or_object(
            module=modstr_parts[0], name=".".join(modstr_parts[1:])
        )


def resolve_lazy_imports(d: dict, *, recursive: bool = True) -> dict:
    """In-place resolves lazy imports in the given dict, recursively.

    .. warning::

        Only recurses on dicts, not on other mutable objects!

    Args:
        d (dict): The dict to resolve lazy imports in
        recursive (bool, optional): Whether to recurse through the dict

    Returns:
        dict: ``d`` but with in-place resolved lazy imports
    """
    for k, v in d.items():
        if isinstance(v, LazyLoader):
            d[k] = v.resolve()

        elif recursive and isinstance(v, dict):
            d[k] = resolve_lazy_imports(d[k], recursive=recursive)

    return d


# -- Misc. --------------------------------------------------------------------


def remove_from_sys_modules(cond: Callable):
    """Removes cached module imports from :py:data:`sys.modules` if their
    fully qualified module name fulfills a certain condition.

    Args:
        cond (Callable): A unary function expecting a single ``str`` argument,
            the module name, e.g. ``numpy.random``. If the function returns
            True, will remove that module.
    """
    for modstr in [m for m in sys.modules if cond(m)]:
        del sys.modules[modstr]


def resolve_types(types: Sequence[Union[type, str]]) -> Sequence[type]:
    """Resolves multiple types, that may be given as module strings, into a
    tuple of types such that it can be used in :py:func:`isinstance` or
    similar functions.

    Args:
        types (Sequence[Union[type, str]]): The types to potentially resolve

    Returns:
        Sequence[type]: The resolved types sequence as a :py:class:`tuple`
    """
    return tuple(t if isinstance(t, type) else import_name(t) for t in types)



"""Takes care of all YAML-related imports and configuration

The ``ruamel.yaml.YAML`` object used here is imported from :py:mod:`paramspace`
and specialized such that it can load and dump dantro classes.
"""

import contextlib
import copy
import io
import logging
import os
from functools import partial as _partial
from typing import Any, Callable, List, Tuple, Union

import matplotlib as mpl
import ruamel.yaml

from .exceptions import raise_improved_exception as _raise_improved_exception

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------

# Import yaml from paramspace and configure it
from paramspace import yaml

yaml.default_flow_style = False

# Register further classes
from ._dag_utils import (
    DAGNode,
    DAGReference,
    DAGTag,
    KeywordArgument,
    Placeholder,
    PositionalArgument,
    ResultPlaceholder,
)

yaml.register_class(Placeholder)
yaml.register_class(ResultPlaceholder)
yaml.register_class(PositionalArgument)
yaml.register_class(KeywordArgument)
yaml.register_class(DAGReference)
yaml.register_class(DAGTag)
yaml.register_class(DAGNode)

# Special constructors ........................................................
# .. For the case of a reference to the "previous" node . . . . . . . . . . . .
yaml.constructor.add_constructor("!dag_prev", lambda l, n: DAGNode(-1))


# .. Colormap and norm creation . . . . . . . . . . . . . . . . . . . . . . . .


def _cmap_constructor(loader, node) -> "matplotlib.colors.Colormap":
    """Constructs a :py:class:`matplotlib.colors.Colormap` object for use in
    plots. Uses the :py:class:`~dantro.plot.utils.color_mngr.ColorManager` and
    directly resolves the colormap object from it.
    """
    from .plot.utils import ColorManager

    if isinstance(node, ruamel.yaml.nodes.MappingNode):
        cmap_kwargs = loader.construct_mapping(node, deep=True)
    else:
        cmap_kwargs = loader.construct_scalar(node)
    cm = ColorManager(cmap=cmap_kwargs)
    cmap = cm.cmap
    cmap._original_yaml = copy.deepcopy(cmap_kwargs)

    return cmap


def _cmap_norm_constructor(loader, node) -> "matplotlib.colors.Colormap":
    """Constructs a :py:class:`matplotlib.colors.Colormap` object for use in
    plots. Uses the :py:class:`~dantro.plot.utils.color_mngr.ColorManager` and
    directly resolves the colormap object from it.
    """
    from .plot.utils import ColorManager

    if isinstance(node, ruamel.yaml.nodes.MappingNode):
        norm_kwargs = loader.construct_mapping(node, deep=True)
    else:
        norm_kwargs = loader.construct_scalar(node)
    cm = ColorManager(norm=norm_kwargs)
    norm = cm.norm
    norm._original_yaml = copy.deepcopy(norm_kwargs)

    return norm


yaml.constructor.add_constructor("!cmap", _cmap_constructor)
yaml.constructor.add_constructor("!cmap_norm", _cmap_norm_constructor)


# Special representers ........................................................

# .. Representers for colormaps and norms . . . . . . . . . . . . . . . . . . .
# If constructed from the !cmap or !cmap_norm tags, these will have the
# additional attribute `_original_yaml`, which is then used for representation.


def _from_original_yaml(representer, node, *, tag: str):
    """For objects where a ``_original_yaml`` attribute was saved."""
    if isinstance(node._original_yaml, dict):
        return representer.represent_mapping(tag, node._original_yaml)
    return representer.represent_scalar(tag, node._original_yaml)


for CmapCls in [
    getattr(mpl.colors, a) for a in dir(mpl.colors) if "Colormap" in a
]:
    yaml.representer.add_representer(
        CmapCls, _partial(_from_original_yaml, tag="!cmap")
    )

for NormCls in [
    getattr(mpl.colors, a) for a in dir(mpl.colors) if "Norm" in a
]:
    yaml.representer.add_representer(
        NormCls, _partial(_from_original_yaml, tag="!cmap_norm")
    )

# -----------------------------------------------------------------------------
# -- Loading and writing ------------------------------------------------------
# -----------------------------------------------------------------------------
# .. Error hints ..............................................................

_YAML_ERROR_HINTS: List[Tuple[Callable, str]] = [
    (
        lambda e: "!dag_prev" in str(e),
        "Did you include a space after the !dag_prev tag in that line?",
    ),
    (
        lambda e: "expected ',' or ']'" in str(e),
        "Did you include a space after the YAML tag defined in that line?",
    ),
    (
        lambda e: True,
        "Read the error message above for details about the error location.",
    ),
]
"""These are evaluated by :py:func:`dantro.exceptions.raise_improved_exception`
and from within :py:func:`.load_yml`.

Entries are of the form ``(match function, hint string)``.
"""

# .............................................................................


def load_yml(
    path: str, *, mode: str = "r", improve_errors: bool = True
) -> Any:
    """Deserializes a YAML file into an object.

    Uses the dantro-internal ``ruamel.yaml.YAML`` object for loading and thus
    supports all registered constructors.

    Args:
        path (str): The path to the YAML file that should be loaded. A ``~`` in
            the path will be expanded to the current user's directory.
        mode (str, optional): Read mode for the file at ``path``
        improve_errors (bool, optional): Whether to improve error messages that
            come from the call to ``yaml.load``. If true, the error message
            is inspected and hints are appended.

    Returns:
        Any: The result of the data loading. Typically, this will be a dict,
            but depending on the structure of the file, it may be some other
            type, including ``None``.
    """
    path = os.path.expanduser(path)
    log.debug("Loading YAML file... mode: %s, path:\n  %s", mode, path)

    with open(path, mode) as yaml_file:
        try:
            return yaml.load(yaml_file)

        except Exception as exc:
            if not improve_errors:
                raise

            # Attempt raising a new and improved error message.
            # Will simply re-raise if that is not the case.
            _raise_improved_exception(exc, hints=_YAML_ERROR_HINTS)


def write_yml(d: Union[dict, Any], *, path: str, mode: str = "w"):
    """Serialize an object using YAML and store it in a file.

    Uses the dantro-internal ``ruamel.yaml.YAML`` object for dumping and thus
    supports all registered representers.

    Args:
        d (dict): The object to serialize and write to file
        path (str): The path to write the YAML output to. A ``~`` in the path
            will be expanded to the current user's directory.
        mode (str, optional): Write mode of the file
    """
    path = os.path.expanduser(path)
    log.debug(
        "Dumping %s to YAML file... mode: %s, path:\n  %s",
        type(d).__name__,
        mode,
        path,
    )

    # Make sure the directory is present
    os.makedirs(os.path.dirname(path), exist_ok=True)

    with open(path, mode) as yaml_file:
        # Add the yaml '---' prefix
        yaml_file.write("---\n")

        # Now dump the rest
        yaml.dump(d, stream=yaml_file)

        # Ensure new line
        yaml_file.write("\n")


# -----------------------------------------------------------------------------
# More specific YAML-based functions


def yaml_dumps(
    obj: Any,
    *,
    register_classes: tuple = (),
    yaml_obj: ruamel.yaml.YAML = None,
    **dump_params,
) -> str:
    """Serializes the given object using a newly created YAML dumper.

    The aim of this function is to provide YAML dumping that is not dependent
    on any package configuration; all parameters can be passed here.

    In other words, his function does _not_ use the dantro._yaml.yaml object
    for dumping but each time creates a new dumper with fixed settings. This
    reduces the chance of interference from elsewhere. Compared to the time
    needed for serialization in itself, the extra time needed to create the
    new ruamel.yaml.YAML object and register the classes is negligible.

    .. note::

        To use dantro's YAML object, it needs to be passed explicitly via the
        ``yaml_obj`` argument! Otherwise a new one will be created which might
        not have the desired classes registered.

    Args:
        obj (Any): The object to dump
        register_classes (tuple, optional): Additional classes to register
        yaml_obj (ruamel.yaml.YAML, optional): If given, use this YAML object
            for dumping. If not given, will create a new one.
        **dump_params: Dumping parameters

    Returns:
        str: The output of serialization

    Raises:
        ValueError: On failure to serialize the given object
    """
    s = io.StringIO()
    if yaml_obj is not None:
        y = yaml_obj
    else:
        y = ruamel.yaml.YAML()

    # Register classes; then apply dumping parameters via object properties
    for Cls in register_classes:
        y.register_class(Cls)

    for k, v in dump_params.items():
        setattr(y, k, v)

    # Serialize
    try:
        y.dump(obj, stream=s)

    except Exception as err:
        raise ValueError(
            f"Could not serialize the given {type(obj)} object!"
        ) from err

    return s.getvalue()



"""Plotting functions that can be used by the
:py:class:`~dantro.plot.creators.pyplot.PyPlotCreator` and derived plot
creators.
"""

from .basic import lineplot
from .generic import errorbars, facet_grid, scatter3d
from .multiplot import multiplot



"""This sub-package implements non-abstract dantro plot creator classes,
based on :py:class:`~dantro.plot.creators.base.BasePlotCreator`"""

from .base import BasePlotCreator
from .psp import MultiversePlotCreator, UniversePlotCreator
from .pyplot import PyPlotCreator

ALL_PLOT_CREATORS = dict(
    base=BasePlotCreator,
    external=PyPlotCreator,  # NOTE Old name, kept for compatibility
    pyplot=PyPlotCreator,
    universe=UniversePlotCreator,
    multiverse=MultiversePlotCreator,
)
"""A mapping of plot creator names to the corresponding types"""



"""This module implements loaders mixin classes for use with the
:py:class:`~dantro.data_mngr.DataManager`.

All these mixin classes should follow the following signature:

.. testcode::

    from dantro.data_loaders import add_loader
    from dantro.base import BaseDataContainer

    class TheTargetContainerClass(BaseDataContainer):
        pass

    class LoadernameLoaderMixin:

        @add_loader(TargetCls=TheTargetContainerClass)
        def _load_loadername(filepath: str, *, TargetCls: type):
            # ...
            return TargetCls(...)

As ensured by the :py:func:`~dantro.data_loaders._registry.add_loader`
decorator, each ``_load_loadername`` method gets supplied with the path to a
file and the ``TargetCls`` argument, which can be called to create an object
of the correct type and name.
In addition, the decorator registers the load function with the dantro
:py:data:`~dantro.data_loaders._registry.DATA_LOADERS` registry, making it
available to :py:class:`~dantro.data_mngr.DataManager` instances that do not
have the mixin added.

By default, and to decouple the loader from the container, it should be
considered to be a static method; in other words: the first positional argument
should ideally *not* be ``self``!
If ``self`` is required for some reason, set the ``omit_self`` option of the
decorator to ``False``, making it a regular (instead of a static) method.
"""

from ._registry import add_loader
from .fspath import FSPathLoaderMixin
from .hdf5 import Hdf5LoaderMixin
from .numpy import NumpyLoaderMixin
from .pandas import PandasLoaderMixin
from .pickle import PickleLoaderMixin
from .text import TextLoaderMixin
from .xarray import XarrayLoaderMixin
from .yaml import YamlLoaderMixin


class AllAvailableLoadersMixin(
    TextLoaderMixin,
    FSPathLoaderMixin,
    YamlLoaderMixin,
    PickleLoaderMixin,
    Hdf5LoaderMixin,
    XarrayLoaderMixin,
    PandasLoaderMixin,
    NumpyLoaderMixin,
):
    """A mixin bundling all data loaders that are available in dantro.
    See the individual mixins for a more detailed documentation.

    If you want all these loaders available in your data manager, inherit from
    this mixin class and :py:class:`~dantro.data_mngr.DataManager`:

    .. testcode::

        import dantro

        class MyDataManager(
            dantro.data_loaders.AllAvailableLoadersMixin,
            dantro.DataManager,
        ):
            pass
    """


# fmt: off

LOADER_BY_FILE_EXT = {
    "txt":      "text",
    "log":      "text",

    "yml":      "yml",
    "yaml":     "yaml",

    "pickle":   "pickle",
    "pkl":      "pkl",

    "hdf5":     "hdf5",
    "h5":       "hdf5",

    "nc":       "xr_dataarray",
    "netcdf":   "xr_dataarray",
    "nc_da":    "xr_dataarray",
    "xrdc":     "xr_dataarray",
    "nc_ds":    "xr_dataset",

    "npy":      "numpy_binary",
    "np_txt":   "numpy_txt",

    "csv":      "pandas_csv",
}
"""A map of file extensions to preferred loader names"""



"""This module implements a :py:class:`dantro.base.BaseDataProxy`
specialization for HDF5 data.
"""

import logging

import numpy as np

from .._import_tools import LazyLoader
from ..base import BaseDataProxy

log = logging.getLogger(__name__)

da = LazyLoader("dask.array")
h5 = LazyLoader("h5py")

# -----------------------------------------------------------------------------


class Hdf5DataProxy(BaseDataProxy):
    """The Hdf5DataProxy is a placeholder for a :py:class:`h5py.Dataset`.

    It saves the filename and dataset name needed to later load the dataset.
    Additionaly, it caches some values that give information on the shape and
    dtype of the dataset, thus further delaying the load to the time the
    *actual* data is required.

    Depending on the type that this proxy is resolved as via the
    :py:meth:`.resolve` method, the corresponding :py:class:`h5py.File` object
    needs to stay open and in memory; it is closed upon garbage-collection of
    this object.
    """

    def __init__(self, obj: "h5py.Dataset", *, resolve_as_dask: bool = False):
        """Initializes a proxy object for a :py:class:`h5py.Dataset` object.

        Args:
            obj (h5py.Dataset): The dataset object to be proxy for
            resolve_as_dask (bool, optional): Whether to resolve the dataset
                object as a delayed :py:class:`dask.array.Array` object, using
                an :py:class:`h5py.Dataset` to initialize it and passing over
                chunk information.
        """
        super().__init__(obj)

        # Information to later resolve the data
        self._fname = obj.file.filename
        self._name = obj.name  # is the full path within the file

        # If file objects need be kept in scope, this is the list to store them
        self._h5files = []

        # Extract some further information from the dataset before, basically
        # all information that can be known without loading the data
        self._shape = obj.shape
        self._dtype = obj.dtype
        self._ndim = obj.ndim
        self._size = obj.size
        self._chunks = obj.chunks

        # Whether to load the hdf5 data through dask.array.from_array
        self._resolve_as_dask = resolve_as_dask

        # Set the tags
        self._tags += ("hdf5",)

        if self._resolve_as_dask:
            self._tags += ("dask",)

    def resolve(self, *, astype: type = None):
        """Resolve the data of this proxy by opening the hdf5 file and loading
        the dataset into a :py:class:`numpy.ndarray` or a type specified by the
        ``astype`` argument.

        Args:
            astype (type, optional): As which type to return the data from the
                dataset this object is proxy for.
                If None, will return as :py:class:`numpy.ndarray`.
                For :py:class:`h5py.Dataset`, the :py:class:`h5py.File` object
                stays in memory until the proxy is deleted.
                Note that if ``resolve_as_dask`` was specified during proxy
                initialization, the data will be loaded as
                :py:class:`dask.array.Array` only if ``astype`` is **not**
                specified in this call!

        Returns:
            type specified by ``astype``: the resolved data.
        """
        if astype is h5.Dataset and not self._resolve_as_dask:
            log.debug(
                "Resolving %s as h5py.Dataset from dataset %s in file "
                "at %s ...",
                self.classname,
                self._name,
                self._fname,
            )

            # Open the file and keep it in scope
            h5file = self._open_h5file()

            # Return the dataset object, which remains valid until the file
            # object is closed, i.e. the proxy goes out of scope
            return h5file[self._name]

        elif astype is None and self._resolve_as_dask:
            log.debug(
                "Resolving %s as h5py.Dataset from dataset %s in file "
                "at %s into a delayed dask array object...",
                self.classname,
                self._name,
                self._fname,
            )

            # Open the file and keep it in scope
            h5file = self._open_h5file()

            # Build the delayed dask array from the h5.Dataset and chunk info
            return da.from_array(h5file[self._name], chunks=self._chunks)

        else:
            # By default, return as numpy array
            astype = astype if astype is not None else np.array

            log.debug(
                "Resolving %s as %s.%s from dataset %s in file at %s ...",
                self.classname,
                astype.__module__,
                astype.__name__,
                self._name,
                self._fname,
            )

            with h5.File(self._fname, "r") as h5file:
                return astype(h5file[self._name])

    # Handling of HDF5 files ..................................................

    def _open_h5file(self) -> "h5py.File":
        """Opens the associated HDF5 file and stores it in ``_h5files`` in
        order to keep it in scope. These file objects are only closed upon
        deletion of this proxy object!

        Returns:
            h5py.File: The newly opened HDF5 file
        """
        h5file = h5.File(self._fname, "r")
        self._h5files.append(h5file)
        return h5file

    def __del__(self):
        """Make sure all potentially still open h5py.File objects are closed"""
        for f in self._h5files:
            try:
                f.close()

            except Exception:
                # Can no longer close it; garbace collection probably already
                # took care of it ... which is fine.
                pass

    # Properties to access information without resolving ......................

    @property
    def shape(self):
        """The cached shape of the dataset, accessible without resolving"""
        return self._shape

    @property
    def dtype(self):
        """The cached dtype of the dataset, accessible without resolving"""
        return self._dtype

    @property
    def ndim(self):
        """The cached ndim of the dataset, accessible without resolving"""
        return self._ndim

    @property
    def size(self):
        """The cached size of the dataset, accessible without resolving"""
        return self._size

    @property
    def chunks(self):
        """The cached chunks of the dataset, accessible without resolving"""
        return self._chunks



"""Implements utilities that revolve around the plotting function which is then
invoked by the :ref:`plot creators <plot_creators>`:

- a decorator to declare a function as a plot function
- the tools to resolve a plotting function from a module or file
"""

import importlib
import importlib.util
import logging
import os
import warnings
from typing import Callable, Sequence, Union

from ..._import_tools import (
    import_module_from_file as _import_module_from_file,
)

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class PlotFuncResolver:
    """Takes care of resolving a plot function"""

    BASE_PKG: str = "dantro.plot.funcs"
    """The default module string to use for relative module imports, where this
    module becomes the base package. Evaluated in :py:meth:`.__init__`.
    """

    def __init__(
        self, *, base_module_file_dir: str = None, base_pkg: str = None
    ):
        """Set up the plot function resolver.

        Args:
            base_module_file_dir (str, optional): If given, ``module_file``
                arguments to :py:meth:`.resolve` that are relative paths will
                be seen relative to this directory. Needs to be an absolute
                directory path and supports ``~`` expansion.
            base_pkg (str, optional): If given, use this base package instead
                for relative module imports instead of :py:attr:`.BASE_PKG`.

        Raises:
            ValueError: If ``base_module_file_dir`` was not absolute
            FileNotFoundError: If ``base_module_file_dir`` is missing or not a
                directory.
        """
        if base_module_file_dir:
            bmfd = os.path.expanduser(base_module_file_dir)

            if not os.path.isabs(bmfd):
                raise ValueError(
                    "Argument `base_module_file_dir` needs to be "
                    f"an absolute path, was not! Got: {bmfd}"
                )

            elif not os.path.exists(bmfd) or not os.path.isdir(bmfd):
                raise FileNotFoundError(
                    "Argument `base_module_file_dir` does not "
                    "exists or does not point to a directory!"
                )

        self.base_module_file_dir = base_module_file_dir
        self.base_pkg = base_pkg if base_pkg else self.BASE_PKG

    def resolve(
        self,
        *,
        plot_func: Union[str, Callable],
        module: str = None,
        module_file: str = None,
    ) -> Callable:
        """Resolve and return the plot function callable

        Args:
            plot_func (Union[str, Callable]): The name or module string of the
                plot function as it can be imported from ``module``. If this is
                a callable will directly return that callable.
            module (str): If ``plot_func`` was the name of the plot
                function, this needs to be the name of the module to import
                that name from.
            module_file (str): Path to the file to load and look for
                the ``plot_func`` in. If ``base_module_file_dir`` is given
                during initialization, this can also be a path relative to that
                directory.

        Returns:
            Callable: The resolved plot function

        Raises:
            TypeError: On bad argument types
        """
        if callable(plot_func):
            log.debug("Received plotting function:  %s", str(plot_func))
            return self._attach_attributes(plot_func)

        elif not isinstance(plot_func, str):
            raise TypeError(
                "Argument `plot_func` needs to be a string or a "
                f"callable, was {type(plot_func)} with value '{plot_func}'."
            )

        # else: need to resolve the module and find the callable in it
        # For less confusing variable names, do some renaming
        plot_func_modstr = plot_func
        del plot_func

        # First resolve the module, either from file or via import
        if module_file:
            mod = self._get_module_from_file(
                module_file, base_module_file_dir=self.base_module_file_dir
            )

        elif isinstance(module, str):
            mod = self._get_module_via_import(
                module=module, base_pkg=self.base_pkg
            )

        else:
            raise TypeError(
                "Could not import a module, because neither argument "
                "`module_file` was given nor did argument `module` have the "
                f"correct type (needs to be string but was {type(module)} "
                f"with value '{module}')."
            )

        # plot_func could be something like "A.B.C.d"; go along the segments to
        # allow for more versatile plot function retrieval
        attr_names = plot_func_modstr.split(".")
        for attr_name in attr_names[:-1]:
            mod = getattr(mod, attr_name)

        # This is now the last module. Get the actual function
        plot_func = getattr(mod, attr_names[-1])

        log.debug("Resolved plotting function:  %s", str(plot_func))

        # Decorate with some attributes, then return the result
        return self._attach_attributes(
            plot_func, module=mod, plot_func_modstr=plot_func_modstr
        )

    def _get_module_from_file(self, path: str, *, base_module_file_dir: str):
        """Returns the module corresponding to the file at the given ``path``.

        This uses :py:func:`~dantro._import_tools.import_module_from_file`
        to carry out the import.
        """
        try:
            return _import_module_from_file(
                path, base_dir=base_module_file_dir
            )
        except ValueError as err:
            raise ValueError(
                "Need to specify `base_module_file_dir` during initialization "
                "to use relative paths for `module_file` argument!"
            ) from err

    def _get_module_via_import(self, *, module: str, base_pkg: str):
        """Returns the module via import.

        Imports ``module`` via importlib, allowing relative imports from the
        package defined as base package.
        """
        return importlib.import_module(module, package=base_pkg)

    def _attach_attributes(
        self,
        plot_func: Callable,
        /,
        *,
        module=None,
        plot_func_modstr: str = None,
    ) -> Callable:
        """Attaches some informational attributes to the plot function."""
        plot_func.modstr = plot_func_modstr
        # TODO attach a `name` here, while the information is available
        return plot_func


# -----------------------------------------------------------------------------


class is_plot_func:
    """This is a decorator class declaring the decorated function as a
    plotting function to use with
    :py:class:`~dantro.plot.creators.base.BasePlotCreator` or derived creators.

    .. note::

        This decorator has a set of specializations that make sense only when
        using a specific creator type!
        For example, the ``helper``-related arguments are only used by
        :py:class:`~dantro.plot.creators.pyplot.PyPlotCreator` and are ignored
        without warning otherwise.
    """

    def __init__(
        self,
        *,
        creator: str = None,
        creator_type: type = None,
        creator_name: str = None,
        use_dag: bool = None,
        required_dag_tags: Sequence[str] = None,
        compute_only_required_dag_tags: bool = True,
        pass_dag_object_along: bool = False,
        unpack_dag_results: bool = False,
        use_helper: bool = None,
        helper_defaults: Union[dict, str] = None,
        supports_animation=False,
        add_attributes: dict = None,
    ):
        """Initialize the decorator.

        .. note::

            Some arguments are only evaluated when using a certain creator
            type, e.g. :py:class:`~dantro.plot.creators.pyplot.PyPlotCreator`.

        Args:
            creator (str, optional): The creator to use; needs to be registered
                with the PlotManager under this name.
            creator_type (type, optional): The type of plot creator to use.
                This argument is DEPRECATED, use ``creator`` instead.
            creator_name (str, optional): The name of the plot creator to use.
                This argument is DEPRECATED, use ``creator`` instead.
            use_dag (bool, optional): Whether to use the data transformation
                framework.
            required_dag_tags (Sequence[str], optional): The DAG tags that are
                required by the plot function.
            compute_only_required_dag_tags (bool, optional): Whether to compute
                only those DAG tags that are specified as required by the plot
                function. This is ignored if no required DAG tags were given
                and can be overwritten by the ``compute_only`` argument.
            pass_dag_object_along (bool, optional): Whether to pass on the DAG
                object to the plot function
            unpack_dag_results (bool, optional): Whether to unpack the results
                of the DAG computation directly into the plot function instead
                of passing it as a dictionary.
            use_helper (bool, optional): Whether to use the
                :py:class:`~dantro.plot.plot_helper.PlotHelper` with this plot.
                Needs :py:class:`~dantro.plot.creators.pyplot.PyPlotCreator`.
                If None, will default to True for supported creators and False
                otherwise.
            helper_defaults (Union[dict, str], optional): Default
                configurations for helpers; these are automatically considered
                to be enabled. If not dict-like, will assume this is an
                absolute path (supporting ``~`` expansion) to a YAML file and
                will load the dict-like configuration from there.
                Needs :py:class:`~dantro.plot.creators.pyplot.PyPlotCreator`.
            supports_animation (bool, optional): Whether the plot function
                supports animation.
                Needs :py:class:`~dantro.plot.creators.pyplot.PyPlotCreator`.
            add_attributes (dict, optional): Additional attributes to add to
                the plot function.

        Raises:
            ValueError: If ``helper_defaults`` was a string but not an absolute
                path.
        """
        from ..._yaml import load_yml

        if helper_defaults and not isinstance(helper_defaults, dict):
            # Interpret as absolute path to yaml file
            fpath = os.path.expanduser(helper_defaults)
            if not os.path.isabs(fpath):
                raise ValueError(
                    "`helper_defaults` string argument was a "
                    f"relative path: {fpath}, but needs to be either a "
                    "dict or an absolute path (~ allowed)."
                )

            log.debug("Loading helper defaults from file %s ...", fpath)
            helper_defaults = load_yml(fpath)

        # Evaluate the creator argument
        if creator:
            if creator_name or creator_type:
                raise ValueError(
                    "Cannot pass both `creator` and `creator_name` or "
                    "`creator_type`!"
                )

        else:
            if creator_name and creator_type:
                raise ValueError(
                    "Cannot pass both of the deprecated decorator arguments "
                    "`creator_name` and `creator_type`. Use `creator` instead."
                )

            _warn_msg = (
                "The `{}` argument is deprecated! Use `creator` instead."
            )
            if creator_name:
                warnings.warn(
                    _warn_msg.format("creator_name"), DeprecationWarning
                )
                creator = creator_name

            elif creator_type:
                warnings.warn(
                    _warn_msg.format("creator_type"), DeprecationWarning
                )
                creator = creator_type

        # Gather those attributes that are to be set as function attributes
        self.pf_attrs = dict(
            is_plot_func=True,
            creator=creator,
            use_dag=use_dag,
            required_dag_tags=required_dag_tags,
            compute_only_required_dag_tags=compute_only_required_dag_tags,
            pass_dag_object_along=pass_dag_object_along,
            use_helper=use_helper,
            helper_defaults=helper_defaults,
            supports_animation=supports_animation,
            **(add_attributes if add_attributes else {}),
        )

    def __call__(self, func: Callable):
        """If there are decorator arguments, __call__() is only called
        once, as part of the decoration process and expects as only argument
        the function to be decorated.
        """
        # Do not actually wrap the function call, but add attributes to it
        for k, v in self.pf_attrs.items():
            setattr(func, k, v)

        # Return the function, now with attributes set
        return func



"""This module implements mixins that provide proxy support"""

import copy
import logging
import warnings
from typing import Union

import numpy as np

from ..abc import AbstractDataProxy

log = logging.getLogger(__name__)

# -----------------------------------------------------------------------------


class ProxySupportMixin:
    """This mixin class overwrites the ``data`` property to allow and resolve
    proxy objects.

    It should be used to add support for certain proxy types to a container.

    A proxy object is a place holder for data that is not yet loaded. It will
    only be loaded if the ``data`` property is directly or indirectly accessed.
    """

    # If the CheckDataMixin is used, it should also allow proxy data
    DATA_ALLOW_PROXY = True
    # NOTE Depending on the order of how the mixins are given, this might
    # not have an effect. Make sure the proxy mixin is loaded after the
    # CheckDataMixin or the container class that uses the CheckDataMixin

    PROXY_RESOLVE_ASTYPE = None
    """Which type to resolve the proxy to"""

    PROXY_RETAIN = False
    """Whether to retain the proxy object after resolving"""

    PROXY_REINSTATE_FAIL_ACTION = "raise"
    """Behaviour upon failure of reinstating a proxy.

    Can be: ``raise``, ``warn``, ``log_warning``, ``log_debug``"""

    PROXY_REINSTATE_FOR_PICKLING = True
    """If true, populates the pickling state with the proxy instead of the data
    """

    # The attribute where a retained proxy is stored (individually for each
    # object, of course).
    _retained_proxy = None

    def __getstate__(self) -> dict:
        """If the data is no longer a proxy, but a proxy was retained, this
        overload adjusts the pickling state such that the proxy object is
        returned instead of the data that was resolved from it. This hels to
        reduce the file size of the pickle.
        """
        if (
            self.data_is_proxy
            or not self.PROXY_REINSTATE_FOR_PICKLING
            or (not self.data_is_proxy and self._retained_proxy is None)
        ):
            return super().__getstate__()

        # else: retrieve the state and adjust it to contain the retained proxy.
        # Need a shallow copy here because state might be __dict__ and its
        # mutability would change the state of self!
        state = copy.copy(super().__getstate__())
        log.debug("Using retained proxy for pickling of %s ...", self.logstr)
        state["_data"] = self._retained_proxy
        state["_retained_proxy"] = None
        return state

    @property
    def data(self):
        """The container data. If the data is a proxy, this call will lead
        to the resolution of the proxy.

        Returns:
            The data stored in this container
        """
        # Have to check whether the data might be a proxy. If so, resolve it.
        if self.data_is_proxy:
            log.debug(
                "Resolving %s for %s ...", self._data.classname, self.logstr
            )

            # Optionally, retain the proxy object. If not doing this, will go
            # out of scope
            if self.PROXY_RETAIN:
                log.debug("...retaining proxy object...")
                self._retained_proxy = self._data

            # Now, overwrite the _data attribute where the proxy used to be
            self._data = self.proxy.resolve(astype=self.PROXY_RESOLVE_ASTYPE)

            # Postprocess the resolved proxy with optional method
            if hasattr(self, "_postprocess_proxy_resolution"):
                log.debug("Calling proxy resolution postprocessing ...")
                self._postprocess_proxy_resolution()

        # Now, the data should be loaded and can be returned
        return self._data

    @property
    def data_is_proxy(self) -> bool:
        """Returns true, if this is proxy data

        Returns:
            bool: Whether the *currently* stored data is a proxy object
        """
        return isinstance(self._data, AbstractDataProxy)

    @property
    def proxy(self) -> Union[AbstractDataProxy, None]:
        """If the data is proxy, returns the proxy data object without using
        the .data attribute (which would trigger resolving the proxy); else
        returns None.

        Returns:
            Union[AbstractDataProxy, None]: If the data is proxy, return the
                proxy object; else None.
        """
        if self.data_is_proxy:
            return self._data
        return self._retained_proxy

    def reinstate_proxy(self):
        """Re-instate a previously retained proxy object, discarding ``_data``."""
        if self.data_is_proxy:
            return

        if self._retained_proxy is None:
            msg = (
                f"Could not reinstate a proxy for {self.logstr} because there "
                "was no proxy retained. Was there one in the first place? "
                "Make sure the `PROXY_RETAIN` class variable is set to True. "
                "To control the behaviour of this message, change the "
                "PROXY_REINSTATE_FAIL_ACTION class variable."
            )
            action = self.PROXY_REINSTATE_FAIL_ACTION

            if action == "raise":
                raise ValueError(msg)

            elif action == "warn":
                warnings.warn(msg, RuntimeWarning)

            elif action in ("log_warning", "log_warn"):
                log.warning(msg)

            elif action == "log_debug":
                log.debug(msg)

            else:
                raise ValueError(
                    "Invalid PROXY_REINSTATE_FAIL_ACTION value "
                    "'{}'! Possible values: raise, warn, "
                    "log_warning, log_debug".format(action)
                )

        else:
            # All good. Reinstate the proxy
            self._data = self._retained_proxy
            log.debug(
                "Reinstated %s for data of %s.",
                self.proxy.classname,
                self.logstr,
            )

    def _format_info(self) -> str:
        """Adds an indicator to whether data is proxy to the info string.
        Additionally, the proxy tags are appended.
        """
        if self.data_is_proxy:
            tags = (
                " ({})".format(", ".join(self.proxy.tags))
                if self.proxy.tags
                else ""
            )
            return f"proxy{tags}, {super()._format_info()}"
        return super()._format_info()


class Hdf5ProxySupportMixin(ProxySupportMixin):
    """Specializes the
    :py:class:`~dantro.mixins.proxy_support.ProxySupportMixin` to the
    capabilities of :py:class:`~dantro.proxy.hdf5.Hdf5DataProxy`, i.e. it
    allows access to the cached properties of the proxy object without
    resolving it.
    """

    @property
    def dtype(self) -> np.dtype:
        """Returns dtype, proxy-aware"""
        if self.data_is_proxy:
            return self.proxy.dtype
        return self.data.dtype

    @property
    def shape(self) -> tuple:
        """Returns shape, proxy-aware"""
        if self.data_is_proxy:
            return self.proxy.shape
        return self.data.shape

    @property
    def ndim(self) -> int:
        """Returns ndim, proxy-aware"""
        if self.data_is_proxy:
            return self.proxy.ndim
        return self.data.ndim

    @property
    def size(self) -> int:
        """Returns size, proxy-aware"""
        if self.data_is_proxy:
            return self.proxy.size
        return self.data.size

    @property
    def chunks(self) -> tuple:
        """Returns chunks, proxy-aware"""
        if self.data_is_proxy:
            return self.proxy.chunks
        return self.data.chunks
